# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

import operator
import sys
import warnings

from typing import Callable, Any
from typing_extensions import Self
from types import FrameType
from dataclasses import dataclass

from caterpillar.exception import StructException
from caterpillar.registry import to_struct
from caterpillar.options import Flag

CTX_PARENT = "_parent"
CTX_OBJECT = "_obj"
CTX_OFFSETS = "_offsets"
CTX_STREAM = "_io"
CTX_FIELD = "_field"
CTX_VALUE = "_value"
CTX_POS = "_pos"
CTX_INDEX = "_index"
CTX_PATH = "_path"
CTX_SEQ = "_is_seq"
CTX_ARCH = "_arch"
CTX_ROOT = "_root"


class Context(dict):
    """Represents a context object with attribute-style access."""

    def __setattr__(self, key: str, value) -> None:
        """
        Sets an attribute in the context.

        :param key: The attribute key.
        :param value: The value to be set.
        """
        self[key] = value

    def __getattribute__(self, key: str):
        """
        Retrieves an attribute from the context.

        :param key: The attribute key.
        :return: The value associated with the key.
        """
        try:
            return object.__getattribute__(self, key)
        except AttributeError:
            return self.__context_getattr__(key)

    def __context_getattr__(self, path: str):
        """
        Retrieves an attribute from the context.

        :param key: The attribute key.
        :return: The value associated with the key.
        """
        nodes = path.split(".")
        obj = (
            self[nodes[0]]
            if nodes[0] in self
            else object.__getattribute__(self, nodes[0])
        )

        for i in range(1, len(nodes)):
            obj = getattr(obj, nodes[i])

        return obj

    def __context_setattr__(self, path: str, value: Any) -> None:
        nodes = path.rsplit(".", 1)
        if len(nodes) == 1:
            self[path] = value
        else:
            obj = self.__context_getattr__(nodes[0])
            setattr(obj, nodes[1], value)

    @property
    def _root(self):
        return self.get("_root", self)


O_CONTEXT_FACTORY = Flag("option.context_factory", value=Context)


class ExprMixin:
    """
    A mixin class providing methods for creating binary and unary expressions.
    """

    def __add__(self, other):
        return BinaryExpression(operator.add, self, other)

    def __sub__(self, other):
        return BinaryExpression(operator.sub, self, other)

    def __mul__(self, other):
        return BinaryExpression(operator.mul, self, other)

    def __floordiv__(self, other):
        return BinaryExpression(operator.floordiv, self, other)

    def __truediv__(self, other):
        return BinaryExpression(operator.truediv, self, other)

    def __mod__(self, other):
        return BinaryExpression(operator.mod, self, other)

    def __pow__(self, other):
        return BinaryExpression(operator.pow, self, other)

    def __xor__(self, other):
        return BinaryExpression(operator.xor, self, other)

    def __and__(self, other):
        return BinaryExpression(operator.and_, self, other)

    def __or__(self, other):
        return BinaryExpression(operator.or_, self, other)

    def __rshift__(self, other):
        return BinaryExpression(operator.rshift, self, other)

    def __lshift__(self, other):
        return BinaryExpression(operator.lshift, self, other)

    __div__ = __truediv__

    def __radd__(self, other):
        return BinaryExpression(operator.add, other, self)

    def __rsub__(self, other):
        return BinaryExpression(operator.sub, other, self)

    def __rmul__(self, other):
        return BinaryExpression(operator.mul, other, self)

    def __rfloordiv__(self, other):
        return BinaryExpression(operator.floordiv, other, self)

    def __rtruediv__(self, other):
        return BinaryExpression(operator.truediv, other, self)

    def __rmod__(self, other):
        return BinaryExpression(operator.mod, other, self)

    def __rpow__(self, other):
        return BinaryExpression(operator.pow, other, self)

    def __rxor__(self, other):
        return BinaryExpression(operator.xor, other, self)

    def __rand__(self, other):
        return BinaryExpression(operator.and_, other, self)

    def __ror__(self, other):
        return BinaryExpression(operator.or_, other, self)

    def __rrshift__(self, other):
        return BinaryExpression(operator.rshift, other, self)

    def __rlshift__(self, other):
        return BinaryExpression(operator.lshift, other, self)

    def __neg__(self):
        return UnaryExpression("neg", operator.neg, self)

    def __pos__(self):
        return UnaryExpression("pos", operator.pos, self)

    def __invert__(self):
        return UnaryExpression("invert", operator.not_, self)

    def __contains__(self, other):
        return BinaryExpression(operator.contains, self, other)

    def __gt__(self, other):
        return BinaryExpression(operator.gt, self, other)

    def __ge__(self, other):
        return BinaryExpression(operator.ge, self, other)

    def __lt__(self, other):
        return BinaryExpression(operator.lt, self, other)

    def __le__(self, other):
        return BinaryExpression(operator.le, self, other)

    def __eq__(self, other):
        return BinaryExpression(operator.eq, self, other)

    def __ne__(self, other):
        return BinaryExpression(operator.ne, self, other)


class ConditionContext:
    """Class implementation of an inline condition.

    Use this class to automatically apply a condition to multiple
    field definitions. Note that this class will only work if it
    has access to the parent stack frame.

    .. code-block:: python

        @struct
        class Format:
            magic: b"MGK"
            length: uint32

            with this.length > 32:
                # other field definitions here
                foo: uint8

    This class will **replace** any existing fields!

    :param condition: a context lambda or constant boolean value
    :type condition: Union[_ContextLambda, bool]
    """

    __slots__ = "func", "annotations", "namelist", "depth"

    def __init__(self, condition, depth=2):
        self.func = condition
        self.annotations = None
        self.namelist = None
        self.depth = depth

    def getframe(self, num: int, msg=None) -> FrameType:
        try:
            return sys._getframe(num)
        except AttributeError as exc:
            raise StructException(msg) from exc

    def __enter__(self) -> Self:
        if (sys.version_info.major, sys.version_info.minor) >= (3, 14):
            warnings.warn(
                "Python3.14 breaks support for Contitional fields. Conditional "
                "statements must be defined manually until a fix has been released."
            )
        frame = self.getframe(self.depth, "Could not enter condition context!")
        # keep track of all annotations
        try:
            self.annotations = frame.f_locals["__annotations__"]
        except AttributeError as exc:
            module = frame.f_locals.get("__module__")
            qualname = frame.f_locals.get("__qualname__")
            msg = f"Could not get annotations in {module} (context={qualname!r})"
            raise StructException(msg) from exc

        # store names before new fields are added
        self.namelist = list(self.annotations)
        return self

    def __exit__(self, *_) -> None:
        # pylint: disable-next=import-outside-toplevel
        from caterpillar.fields import Field

        new_names = set(self.annotations) - set(self.namelist)
        for name in new_names:
            # modify newly created fields
            field = self.annotations[name]
            if isinstance(field, Field):
                # field already defined/created -> check for condition
                if field.has_condition():
                    # the field's condition AND this one must be true
                    field.condition = BinaryExpression(
                        operator.and_, field.condition, self.func
                    )
                else:
                    field //= self.func
            else:
                # create a field (other attributes will be modified later)

                # ISSUE #15: The annotation must be converted to a _StructLike
                # object. In case we have struct classes, the special __struct__
                # attribute must be used.
                struct_obj = to_struct(field)
                if not isinstance(struct_obj, Field):
                    struct_obj = Field(struct_obj)

                struct_obj.condition = self.func
                self.annotations[name] = struct_obj

        self.annotations = None
        self.namelist = None


@dataclass(repr=False)
class BinaryExpression(ExprMixin):
    """
    Represents a binary expression.

    :param operand: The binary operator function.
    :param left: The left operand.
    :param right: The right operand.
    """

    operand: Callable[[Any, Any], Any]
    left: Any
    right: Any

    def __call__(self, context: Context, **kwds):
        lhs = self.left(context, **kwds) if callable(self.left) else self.left
        rhs = self.right(context, **kwds) if callable(self.right) else self.right
        return self.operand(lhs, rhs)

    def __repr__(self) -> str:
        return f"{self.operand.__name__}{{{self.left!r}, {self.right!r}}}"

    def __enter__(self):
        # pylint: disable-next=attribute-defined-outside-init
        self._cond = ConditionContext(self, depth=3)
        self._cond.__enter__()
        return self

    def __exit__(self, *_):
        self._cond.__exit__(*_)


@dataclass
class UnaryExpression:
    """
    Represents a unary expression.

    :param name: The name of the unary operator.
    :param operand: The unary operator function.
    :param value: The operand.
    """

    name: str
    operand: Callable[[Any], Any]
    value: Any

    def __call__(self, context, **kwds):
        value = self.value(context, **kwds) if callable(self.value) else self.value
        return self.operand(value)

    def __repr__(self) -> str:
        return f"{self.operand.__name__}{{{self.value!r}}}"

    def __enter__(self):
        # pylint: disable-next=attribute-defined-outside-init
        self._cond = ConditionContext(self, depth=3)
        self._cond.__enter__()
        return self

    def __exit__(self, *_):
        self._cond.__exit__(*_)


class ContextPath(ExprMixin):
    """
    Represents a lambda function for retrieving a value from a Context based on a specified path.
    """

    def __init__(self, path=None) -> None:
        """
        Initializes a ContextPath instance with an optional path.

        :param path: The path to use when retrieving a value from a Context.
        """
        self.path = path
        self._ops_ = []
        self.call_kwargs = None
        self.getitem_args = None

    def __call__(self, context=None, **kwds):
        """
        Calls the lambda function to retrieve a value from a Context.

        :param context: The Context from which to retrieve the value.
        :param kwds: Additional keyword arguments.
        :return: The value retrieved from the Context based on the path.
        """
        if context is None:
            self._ops_.append((operator.call, (), kwds))
            return self
        value = context.__context_getattr__(self.path)
        for operation, args, kwargs in self._ops_:
            value = operation(value, *args, **kwargs)
        return value

    def __getitem__(self, key) -> Self:
        self._ops_.append((operator.getitem, (key,), {}))
        return self

    def __type__(self) -> type:
        return object

    def __getattribute__(self, key: str) -> ContextPath:
        """
        Gets an attribute from the ContextPath, creating a new instance if needed.

        :param key: The attribute key.
        :return: A new ContextPath instance with an updated path.
        """
        try:
            return super().__getattribute__(key)
        except AttributeError:
            if not self.path:
                return ContextPath(key)
            return ContextPath(".".join([self.path, key]))

    def __repr__(self) -> str:
        """
        Returns a string representation of the ContextPath.

        :return: A string representation.
        """
        extra = []
        for operation, args, kwargs in self._ops_:
            data = []
            if len(args) > 0:
                data.append(*map(repr, args))
            if len(kwargs) > 0:
                data.append(*[f"{x}={y!r}" for x, y in kwargs.items()])
            extra.append(f"{operation.__name__}({', '.join(data)})")

        if len(extra) == 0:
            return f"Path({self.path!r})"

        return f"Path({self.path!r}, {', '.join(extra)})"

    def __str__(self) -> str:
        """
        Returns a string representation of the path.

        :return: A string representation of the path.
        """
        return self.path

    @property
    def parent(self) -> ContextPath:
        path = f"{CTX_PARENT}.{CTX_OBJECT}"
        if not self.path:
            return ContextPath(path)
        return ContextPath(".".join([self.path, path]))


class ContextLength(ExprMixin):
    def __init__(self, path: ContextPath) -> None:
        self.path = path

    def __call__(self, context=None, **kwds):
        """
        Calls the lambda function to retrieve a value from a Context.

        :param context: The Context from which to retrieve the value.
        :param kwds: Additional keyword arguments (ignored in this implementation).
        :return: The value retrieved from the Context based on the path.
        """
        return len(self.path(context))

    def __repr__(self) -> str:
        return f"len({self.path!r})"


this = ContextPath(CTX_OBJECT)
ctx = ContextPath()
parent = ContextPath(".".join([CTX_PARENT, CTX_OBJECT]))
root = ContextPath(CTX_ROOT)


# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any

from caterpillar.abc import _StructLike
from caterpillar.exception import ValidationError


class TypeConverter:
    """
    A utility class to define and manage type conversion handlers for annotations.

    Type converters take the placed annotation and convert it into
    a :code:`_StructLike` object. For instance, a simple converter that
    always returns :code:`uint32` for an object of the Python `int` class might look
    like this:

    >>> tc = TypeConverter(int, lambda a, _: Const(a, uint16))

    or directly as annotation

    >>> @TypeConverter(int)
    ... def int_converter(annotation, kwargs):
    ...     return Const(annotation, uint16)
    ...

    :param target: the target type, defaults to None
    :type target: Type, optional
    :param delegate: optional delegation function, defaults to None
    :type delegate: Callable[[Any, dict], _StructLike], optional
    """

    def __init__(self, target=None, delegate=None) -> None:
        self.target = target
        self.delegate = delegate

    def matches(self, annotation: Any) -> bool:
        """Check if this converter matches the given annotation type.

        :param annotation: he type or object to match against the target.
        :type annotation: Any
        :return: True if the annotation matches the target type; otherwise False.
        :rtype: bool
        """
        return False if self.target is None else isinstance(annotation, self.target)

    def convert(self, annotation: Any, kwargs: dict) -> _StructLike:
        """Convert the given annotation (uses the delegate function by default).

        :param annotation: The object to be converted.
        :type annotation: Any
        :param kwargs: Additional arguments for the conversion.
        :type kwargs: dict
        :raises NotImplementedError: If no delegate is defined for the converter.
        :return: The converted object.
        :rtype: _StructLike
        """
        if self.delegate is None:
            raise NotImplementedError("NOT IMPLEMENTED")

        return self.delegate(annotation, kwargs)

    def __call__(self, delegate) -> Any:
        """Sets the delegate function for this converter and return the instance."""
        self.delegate = delegate
        return self

    def __repr__(self) -> str:
        """Returns a string representation of the TypeConverter instance."""
        if not self.target:
            return f"<{self.__class__.__name__}>"

        return f"<{self.__class__.__name__} for {self.target.__name__}>"


# A global registry to store type converters.
annotation_registry: list[TypeConverter] = []


def to_struct(obj: Any, **kwargs) -> _StructLike:
    """Convert an object to a :code:`_StruckLike` object using registered type converters.

    This function will not convert any objects that are already
    implementing the functions of :code:`_StructLike`.

    :param obj: The object to be converted.
    :type obj: Any
    :param kwargs: Additional arguments passed to the conversion handler.
    :raises ValidationError: If no matching type converter is found for the object.
    :return: the converted object
    :rtype: _StructLike
    """

    # If the object is already a structured-like instance, return it.
    if isinstance(obj, _StructLike):
        return obj

    global annotation_registry

    # Find and use only the first registered converter that matches the object's type.
    handler = next(filter(lambda x: x.matches(obj), annotation_registry), None)
    if handler is None:
        msg = (
            f"The object of type '{obj.__class__}' could not be converted to a struct, because "
            "the placed annotation does not have a corresponding handler:\n "
            f"type: {type(obj)},\n annotation: {obj!r}"
        )
        raise ValidationError(msg)

    return handler.convert(obj, kwargs)



from typing import Callable, Iterable, Type, TypeVar, overload
from caterpillar._C import * # noqa
from caterpillar._C import __all__ as _c_all

_T = TypeVar("_T")

@overload
def struct(
    cls: None = None,
    /,
    options: Iterable[Option] = ...,
    endian: Endian = ...,
    arch: Arch = ...,
    field_options: Iterable[Option] = ...,
) -> Callable[[Type[_T]], Type[_T]]: ...
@overload
def struct(
    cls: Type[_T],
    /,
    options: Iterable[Option] = ...,
    endian: Endian = ...,
    arch: Arch = ...,
    field_options: Iterable[Option] = ...,
) -> Type[_T]: ...


# pyright: reportUnsupportedDunderAll=false
__all__ = _c_all + ["struct"]



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from ._base import Field, INVALID_DEFAULT, DEFAULT_OPTION, singleton
from ._mixin import FieldMixin, FieldStruct, Chain, Operator, get_args, get_kwargs
from .common import (
    PyStructFormattedField,
    Transformer,
    Const,
    ConstBytes,
    ConstString,
    Enum,
    String,
    Bytes,
    Memory,
    Computed,
    Pass,
    CString,
    Prefixed,
    Int,
    UInt,
    padding,
    char,
    boolean,
    int8,
    uint8,
    int16,
    uint16,
    int24,
    uint24,
    int32,
    uint32,
    int64,
    uint64,
    ssize_t,
    size_t,
    float16,
    float32,
    float64,
    double,
    void_ptr,
    Uuid,
    Aligned,
    align,
    Lazy,
    ENUM_STRICT,
)
from .varint import VarInt, VARINT_LSB, vint
from .compression import (
    Compressed,
    ZLibCompressed,
    Bz2Compressed,
    LZMACompressed,
    LZOCompressed,
)
from .crypto import Encrypted, Xor, Or, And, KeyCipher
from .net import MAC, MACAddress, IPv4Address, IPv6Address
from .pointer import (
    uintptr,
    intptr,
    offintptr,
    offuintptr,
    Pointer,
    pointer,
    intptr_fn,
    PTR_STRICT,
    relative_pointer,
    RelativePointer,
    uintptr_fn,
)
from .conditional import ConditionalChain, If, Else, ElseIf
from .hook import IOHook
from .digest import (
    Digest,
    Algorithm,
    Md5,
    Sha1,
    Sha2_256,
    Sha2_224,
    Sha2_384,
    Sha2_512,
    Sha3_224,
    Sha3_256,
    Sha3_384,
    Sha3_512,
    Crc32,
    Adler,
    HMAC,
    DigestField,
    DigestFieldAction,
    Md5_Algo,
    Md5_Field,
    Sha1_Algo,
    Sha1_Field,
    Sha2_256_Algo,
    Sha2_256_Field,
    Sha2_224_Algo,
    Sha2_224_Field,
    Sha2_384_Algo,
    Sha2_384_Field,
    Sha2_512_Algo,
    Sha2_512_Field,
    Sha3_224_Algo,
    Sha3_224_Field,
    Sha3_256_Algo,
    Sha3_256_Field,
    Sha3_384_Algo,
    Sha3_384_Field,
    Sha3_512_Algo,
    Sha3_512_Field,
    Crc32_Algo,
    Crc32_Field,
    Adler_Algo,
    Adler_Field,
    CTX_DIGEST,
    CTX_DIGEST_ALGO,
    CTX_DIGEST_HOOK,
    CTX_DIGEST_OBJ,
    HMACAlgorithm,
)

__all__ = [
    "Digest",
    "Algorithm",
    "Md5",
    "Sha1",
    "Sha2_256",
    "Sha2_224",
    "Sha2_384",
    "Sha2_512",
    "Sha3_224",
    "Sha3_256",
    "Sha3_384",
    "Sha3_512",
    "Crc32",
    "Adler",
    "HMAC",
    "DigestField",
    "DigestFieldAction",
    "Md5_Algo",
    "Md5_Field",
    "Sha1_Algo",
    "Sha1_Field",
    "Sha2_256_Algo",
    "Sha2_256_Field",
    "Sha2_224_Algo",
    "Sha2_224_Field",
    "Sha2_384_Algo",
    "Sha2_384_Field",
    "Sha2_512_Algo",
    "Sha2_512_Field",
    "Sha3_224_Algo",
    "Sha3_224_Field",
    "Sha3_256_Algo",
    "Sha3_256_Field",
    "Sha3_384_Algo",
    "Sha3_384_Field",
    "Sha3_512_Algo",
    "Sha3_512_Field",
    "Crc32_Algo",
    "Crc32_Field",
    "Adler_Algo",
    "Adler_Field",
    "CTX_DIGEST",
    "CTX_DIGEST_ALGO",
    "CTX_DIGEST_HOOK",
    "CTX_DIGEST_OBJ",
    "HMACAlgorithm",
    "uintptr",
    "intptr",
    "offintptr",
    "offuintptr",
    "Pointer",
    "pointer",
    "intptr_fn",
    "PTR_STRICT",
    "relative_pointer",
    "RelativePointer",
    "uintptr_fn",
    "Compressed",
    "ZLibCompressed",
    "Bz2Compressed",
    "LZMACompressed",
    "LZOCompressed",
    "PyStructFormattedField",
    "Transformer",
    "Const",
    "ConstBytes",
    "ConstString",
    "Enum",
    "String",
    "Bytes",
    "Memory",
    "Computed",
    "Pass",
    "CString",
    "Prefixed",
    "Int",
    "UInt",
    "padding",
    "char",
    "boolean",
    "int8",
    "uint8",
    "int16",
    "uint16",
    "int24",
    "uint24",
    "int32",
    "uint32",
    "int64",
    "uint64",
    "ssize_t",
    "size_t",
    "float16",
    "float32",
    "float64",
    "double",
    "void_ptr",
    "Uuid",
    "Aligned",
    "align",
    "Lazy",
    "ENUM_STRICT",
    "Field",
    "INVALID_DEFAULT",
    "DEFAULT_OPTION",
    "singleton",
    "FieldMixin",
    "FieldStruct",
    "Chain",
    "Operator",
    "get_args",
    "get_kwargs",
    "VarInt",
    "VARINT_LSB",
    "vint",
    "Encrypted",
    "Xor",
    "Or",
    "And",
    "KeyCipher",
    "MAC",
    "MACAddress",
    "IPv4Address",
    "IPv6Address",
    "ConditionalChain",
    "If",
    "Else",
    "ElseIf",
    "IOHook",
]



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass
from typing import Any


@dataclass(init=False, eq=False)
class Flag:
    """Simple customizable user-flag."""

    name: str
    """The name of this flag"""

    value: Any = None
    """
    Optional configuration value.
    """

    def __init__(self, name: str, value=None) -> None:
        self.name = name
        self.value = value
        self._hash_ = hash(name)

    def __hash__(self) -> int:
        """
        Custom hash method based on the flag's name.

        :return: The hash value.
        """
        return self._hash_

    def __eq__(self, value: object, /) -> bool:
        # custom check to allow c_Option objects to be used
        return getattr(value, "name", None) == self.name


#: Defaults that will be applied to **all** structs.
GLOBAL_STRUCT_OPTIONS = set()

#: Defaults that will be applied on **all** unions.
GLOBAL_UNION_OPTIONS = set()

#: Default field flags that will be applied on **all** fields.
GLOBAL_FIELD_FLAGS = set()

#: Default field flags that will be applied on **all** bit-fields.
GLOBAL_BITFIELD_FLAGS = set()


def configure(base, *flags: Flag) -> None:
    """
    Update the base set of flags with additional flags.

    :param base: The base set of flags.
    :param flags: Additional flags to be added.
    """
    base.update(flags)


def set_struct_flags(*flags: Flag, with_union: bool = False) -> None:
    """
    Set default flags for structs.

    :param flags: Flags to be set.
    :param with_union: Whether to include the flags for unions as well.
    """
    configure(GLOBAL_STRUCT_OPTIONS, *flags)
    if with_union:
        configure(GLOBAL_UNION_OPTIONS, *flags)


def set_field_flags(*flags: Flag) -> None:
    """
    Set default flags for fields.

    :param flags: Flags to be set.
    """
    configure(GLOBAL_FIELD_FLAGS, *flags)


def set_union_flags(*flags: Flag) -> None:
    """
    Set default flags for unions.

    :param flags: Flags to be set.
    """
    configure(GLOBAL_UNION_OPTIONS, *flags)


def get_flags(obj, attr=None):
    """
    Get the flags associated with an object.

    :param obj: The object.
    :param attr: The attribute name containing the flags (default is "flags").
    :return: The set of flags, or None if no flags are present.
    """
    return getattr(obj, attr or "flags", None)


def has_flag(flag, obj, attr=None) -> bool:
    """
    Check if an object has a specific flag.

    :param flag: The flag to check, either as a string or a Flag instance.
    :param obj: The object.
    :param attr: The attribute name containing the flags (default is "flags").
    :return: True if the flag is present, False otherwise.
    """
    flags = get_flags(obj, attr)
    if not flags or not isinstance(flags, set):
        return False

    if isinstance(flag, str):
        flag = Flag(flag)
    return flag in flags


def get_flag(name: str, obj, attr=None):
    """
    Get a specific flag associated with an object.

    :param name: The name of the flag.
    :param obj: The object.
    :param attr: The attribute name containing the flags (default is "flags").
    :return: The Flag instance if found, otherwise None.
    """
    flags = get_flags(obj, attr)
    for flag in flags or []:
        if flag.name == name:
            return flag


###############################################################################
# Default defined options
###############################################################################

# for structs and unions
S_DISCARD_UNNAMED = Flag("struct.discard_unnamed")
S_DISCARD_CONST = Flag("struct.discard_const")
S_UNION = Flag("struct.union")
S_REPLACE_TYPES = Flag("struct.replace_types")
S_EVAL_ANNOTATIONS = Flag("struct.eval_annotations")
S_ADD_BYTES = Flag("struct.bytes_method")
S_SLOTS = Flag("struct.slots")

# for fields
F_KEEP_POSITION = Flag("field.keep_position")
F_DYNAMIC = Flag("field.dynamic")
F_SEQUENTIAL = Flag("field.sequential")
F_OFFSET_OVERRIDE = Flag("field.offset_override")

# value intentionally left blank
O_ARRAY_FACTORY = Flag("option.array_factory", value=None)

# bitfield options
B_OVERWRITE_ALIGNMENT = Flag("bitfield.overwrite_alignment")
B_GROUP_END = Flag("bitfield.group.end")
B_GROUP_NEW = Flag("bitfield.group.new")
B_GROUP_KEEP = Flag("bitfield.group.keep")
B_NO_AUTO_BOOL = Flag("bitfield.no_auto_bool")



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Callable, Optional
from typing_extensions import Buffer
from caterpillar.abc import _ContextLike, _ContextLambda
from caterpillar.context import CTX_STREAM as CTX_STREAM
from io import RawIOBase

HookInit = _ContextLambda[None]
HookUpdate = Callable[[bytes, _ContextLike], Optional[bytes]]
HookRead = Callable[[bytes, _ContextLike], Optional[bytes]]
HookWrite = Callable[[bytes, _ContextLike], Optional[bytes]]
HookFinish = _ContextLambda[None]

class IOHook(RawIOBase):
    def __init__(
        self,
        io: RawIOBase | None,
        init: HookInit | None = None,
        update: HookUpdate | None = None,
        read: HookRead | None = None,
        write: HookWrite | None = None,
        finish: HookFinish | None = None,
    ) -> None: ...
    def assert_context_set(self) -> None: ...
    def init(self, context: _ContextLike) -> None: ...
    def finish(self, context: _ContextLike) -> None: ...
    def seekable(self) -> bool: ...
    def readable(self) -> bool: ...
    def read(self, size: int = -1) -> bytes | None: ...
    def write(self, b: Buffer, /) -> int | None: ...
    def writable(self) -> bool: ...
    def tell(self) -> int: ...
    def seek(self, offset: int, whence: int = 0) -> int: ...



from mrcrowbar.fields import Bits8, UInt24_LE, UInt32_LE, BlockField, UInt8
from mrcrowbar.models import Block, CString, CStringN
from mrcrowbar.refs import Ref


class Entry(Block):
    num1 = UInt8()
    num2 = UInt24_LE()
    bool1 = Bits8(0x0004, bits=0b10000000, endian="little")
    num4 = Bits8(0x0004, bits=0b01110000, endian="little")
    fixedarray1 = UInt8(0x0005, count=3)
    name1 = CString(0x0008)
    name2 = CString(length_field=UInt8)


class Format(Block):
    count = UInt32_LE(0x0000)
    entries = BlockField(Entry, 0x0004, count=Ref("count"))


if __name__ == "__main__":
    import sys
    import timeit

    try:
        from rich import print
    except ImportError:
        pass

    with open(sys.argv[1], "rb") as fp:
        data = fp.read()

    obj = Format(data)
    time = timeit.timeit(lambda: Format(data), number=1000) / 1000
    print("[bold]Parsing measurements:[/]")
    print(f"[bold]default[/]  {time:.10f} sec/call\n")

    time = timeit.timeit(obj.export_data, number=1000) / 1000
    print("[bold]Building measurements:[/]")
    print(f"[bold]default[/]  {time:.10f} sec/call")



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from abc import abstractmethod
from io import IOBase
from typing import Protocol, runtime_checkable
from typing import Any, Optional, Union, Callable

#: Type alias for IOBase to indicate a stream type
_StreamType = IOBase
_StreamFactory = Callable[[], _StreamType]

_GreedyType = type(...)
_PrefixedType = slice


@runtime_checkable
class _ContextLike(Protocol):
    """
    A dictionary-like object used as a context for packing and unpacking.

    It allows direct attribute access and modification.
    """

    def __context_getattr__(self, path: str) -> Any:
        pass

    def __context_setattr__(self, path: str, value: Any) -> None:
        pass

    @property
    def _root(self):
        pass

    def __getitem__(self, key: str) -> Any:
        pass

    def __setitem__(self, key: str, value: Any) -> None:
        pass

    def get(self, key: str, default: Any | None = None) -> Any:
        pass


@runtime_checkable
class _ContextLambda(Protocol):
    """
    An abstract base class for a lambda function that takes a _ContextLike
    object as an argument.
    """

    def __call__(self, context: _ContextLike) -> Any:
        pass


@runtime_checkable
class _SupportsActionUnpack(Protocol):
    def __action_pack__(self, context: _ContextLike) -> None:
        pass


@runtime_checkable
class _SupportsActionPack(Protocol):
    def __action_unpack__(self, context: _ContextLike) -> None:
        pass


@runtime_checkable
class _SupportsPack(Protocol):
    """
    An abstract base class for objects that support packing data into a binary stream.
    """

    def __pack__(self, obj: Any, context: _ContextLike) -> None:
        pass


@runtime_checkable
class _SupportsUnpack(Protocol):
    """
    An abstract base class for objects that support unpacking data from a binary stream.
    """

    def __unpack__(self, context: _ContextLike) -> Any:
        pass


@runtime_checkable
class _SupportsSize(Protocol):
    """
    An abstract base class for objects that support determining the size of packed data.
    """

    def __size__(self, context: _ContextLike) -> int:
        return 0


@runtime_checkable
class _StructLike(Protocol):
    """
    An abstract base class for struct-like objects that can be packed, unpacked, and have a size.
    """

    def __size__(self, context: _ContextLike) -> int:
        return 0

    def __unpack__(self, context: _ContextLike) -> Any:
        pass

    def __pack__(self, obj: Any, context: _ContextLike) -> None:
        pass


@runtime_checkable
class _SupportsType(Protocol):
    def __type__(self) -> Optional[Union[type, str]]:
        pass


class _ContainsStruct(Protocol):
    """
    An abstract base class indicating that a class contains a _StructLike object.
    """

    __struct__: _StructLike


@runtime_checkable
class _SwitchLike(Protocol):
    """
    An abstract base class for a switch-like object that generates a _StructLike
    based on a value and context.
    """

    @abstractmethod
    def __call__(self, value: Any, context: _ContextLike, **kwds) -> _StructLike:
        pass


@runtime_checkable
class _SupportsBits(Protocol):
    def __bits__(self) -> int: ...


@runtime_checkable
class _ContainsBits(Protocol):
    __bits__: int


__all__ = [
    "_ContextLike",
    "_ContextLambda",
    "_SupportsActionUnpack",
    "_SupportsActionPack",
    "_SupportsPack",
    "_SupportsUnpack",
    "_SupportsSize",
    "_StructLike",
    "_SupportsType",
    "_ContainsStruct",
    "_SwitchLike",
    "_SupportsBits",
    "_ContainsBits",
]



# type: ignore
from caterpillar.py import (
    struct,
    set_struct_flags,
    S_REPLACE_TYPES,
    uint8,
    uint16,
    TemplateTypeVar,
    template,
    derive,
    pack,
    this,
)

try:
    from rich import print
except ImportError:
    pass

set_struct_flags(S_REPLACE_TYPES)


@struct
class BaseFormat:
    """Default class documentation"""

    #: inline comment
    f1: uint8


A = TemplateTypeVar("A")
B = TemplateTypeVar("B")


@template(A, B)
class FormatTemplate(BaseFormat):
    """Template class doc-comment"""

    f2: A[this.f1]
    """Template field doc-comment"""

    #: inline template field comment
    f3: B


#: anonymous generated partial template
Format8 = derive(FormatTemplate, uint8, partial=True)


@struct
class Format(derive(Format8, B=uint8)):
    #: inline comment
    f4: uint8


#: inline data comment
Format16 = derive(FormatTemplate, uint16, uint16, name=...)

if __name__ == "__main__":
    # Format(f1: int, f2: List, f3: int, f4: int)
    print(Format.__doc__)
    # Format16(f1: int, f2: List, f3: int)
    print(Format16.__doc__)

    obj = Format(2, [3, 4], 0xEE, 0xFF)
    # b'\x02\x03\x04\xee\xff'
    print(pack(obj))

    obj = Format16(2, [3, 4], 0xFF)
    # b'\x02\x03\x00\x04\x00\xff\x00'
    print(pack(obj))



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Protocol, runtime_checkable

from caterpillar.abc import _StructLike
from caterpillar.shared import getstruct, hasstruct
from caterpillar.fields._mixin import get_kwargs
from caterpillar.fields.common import Transformer, Bytes


@runtime_checkable
class _Compressor(Protocol):
    """
    A protocol for compression algorithms, requiring two methods:
    - compress: to compress data
    - decompress: to decompress data

    Any class implementing this protocol must define these methods.
    """

    def compress(self, data: bytes, **kwds) -> bytes:
        """
        Compress the provided data.

        :param data: The data to be compressed.
        :type data: bytes
        :param kwds: Additional keyword arguments specific to the compression method.
        :return: The compressed data.
        :rtype: bytes
        """
        return b""

    def decompress(self, data: bytes, **kwds) -> bytes:
        """
        Decompress the provided data.

        :param data: The compressed data to be decompressed.
        :type data: bytes
        :param kwds: Additional keyword arguments specific to the decompression method.
        :return: The decompressed data.
        :rtype: bytes
        """
        return b""


class Compressed(Transformer):
    """
    A transformer class for handling compression and decompression of data.

    This class takes a compression algorithm that implements the `_Compressor` protocol and
    applies it to the data for compression or decompression operations. It allows for
    additional customization through the `comp_kwargs` and `decomp_kwargs` parameters.

    :param compressor: The compression algorithm that implements the _Compressor protocol.
    :type compressor: _Compressor
    :param struct: The struct to be compressed or decompressed, should implement _ContainsStruct or _StructLike.
    :type struct: Union[_ContainsStruct, _StructLike]
    :param comp_kwargs: Optional keyword arguments for the compression method.
    :type comp_kwargs: Optional[dict]
    :param decomp_kwargs: Optional keyword arguments for the decompression method.
    :type decomp_kwargs: Optional[dict]
    """

    def __init__(
        self,
        compressor,
        struct,
        comp_kwargs=None,
        decomp_kwargs=None,
    ) -> None:
        if hasstruct(struct):
            struct = getstruct(struct)
        super().__init__(struct)
        self.compressor = compressor
        self.comp_args = comp_kwargs or {}
        self.decomp_args = decomp_kwargs or {}

    def encode(self, obj: bytes, context) -> bytes:
        """
        Compress the input data using the provided compressor.

        :param obj: The data to be compressed.
        :type obj: bytes
        :param context: Context information for compression (e.g., field-specific metadata).
        :type context: _ContextLike
        :return: The compressed data.
        :rtype: bytes
        """
        return self.compressor.compress(obj, **get_kwargs(self.comp_args, context))

    def decode(self, parsed: bytes, context) -> bytes:
        """
        Decompress the input data using the provided compressor.

        :param parsed: The compressed data to be decompressed.
        :type parsed: bytes
        :param context: Context information for decompression (e.g., field-specific metadata).
        :type context: _ContextLike
        :return: The decompressed data.
        :rtype: bytes
        """
        return self.compressor.decompress(
            parsed, **get_kwargs(self.decomp_args, context)
        )


def compressed(
    lib,
    obj,
    comp_kwargs=None,
    decomp_kwargs=None,
) -> _StructLike:
    if callable(obj) or isinstance(obj, int) or obj is ...:
        obj = Bytes(obj)
    return Compressed(lib, obj, comp_kwargs, decomp_kwargs)


def ZLibCompressed(
    obj,
    comp_kwargs=None,
    decomp_kwargs=None,
):
    """
    Create a struct representing zlib compression.
    """
    try:
        import zlib

        return compressed(zlib, obj, comp_kwargs, decomp_kwargs)
    except ImportError:
        raise NotImplementedError("Could not import zlib!")


def Bz2Compressed(
    obj,
    comp_kwargs=None,
    decomp_kwargs=None,
):
    """
    Create a struct representing bz2 compression.
    """
    try:
        import bz2

        return compressed(bz2, obj, comp_kwargs, decomp_kwargs)
    except ImportError:
        raise NotImplementedError("Could not import bz2!")


def LZMACompressed(
    obj,
    comp_kwargs=None,
    decomp_kwargs=None,
):
    """
    Create a struct representing lzma compression.
    """
    try:
        import lzma

        return compressed(lzma, obj, comp_kwargs, decomp_kwargs)
    except ImportError:
        raise NotImplementedError("Could not import lzma!")


def LZOCompressed(
    obj,
    comp_kwargs=None,
    decomp_kwargs=None,
):
    """
    Create a struct representing LZO compression.
    """
    try:
        # install package manuall with pip install lzallright
        import lzallright

        return compressed(lzallright.LZOCompressor(), obj, comp_kwargs, decomp_kwargs)
    except ImportError:
        raise NotImplementedError("Could not import lzallright!")



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from collections.abc import Iterable
from caterpillar.fields.common import Int
from caterpillar.shared import (
    ATTR_ACTION_PACK,
    ATTR_ACTION_UNPACK,
    ATTR_BYTEORDER,
    typeof,
    ATTR_BITS,
    ATTR_SIGNED,
)
from caterpillar.byteorder import (
    LittleEndian,
    SysNative,
    system_arch,
)
from caterpillar.options import (
    B_GROUP_NEW,
    GLOBAL_BITFIELD_FLAGS,
    GLOBAL_STRUCT_OPTIONS,
    GLOBAL_UNION_OPTIONS,
    B_OVERWRITE_ALIGNMENT,
    B_GROUP_END,
    B_GROUP_KEEP,
    B_NO_AUTO_BOOL,
    Flag,
)
from caterpillar.fields import (
    Field,
    Pass,
    INVALID_DEFAULT,
)
from caterpillar.exception import StructException
from caterpillar.context import CTX_PATH, O_CONTEXT_FACTORY, CTX_OBJECT, CTX_STREAM

from ._struct import Struct, sizeof


# --- Bitfield Concept ---
# NEW REVISED CONCEPT
# Each Bitfield instance maintains a sequence of bitfield groups, where each group
# contains a collection of sized fields. A bitfield group may consist of either multiple
# entries (i.e., any types that can be converted to an integral type) or a single
# _StructLike object. For example, consider the following bitfield definition:
#
#   @bitfield
#   class Format:
#       a1: 1
#       a2: 1 - boolean
#       _ : 0
#       b1: char
#       c1: uint32
#
# This Bitfield definition will generate three distinct bitfield groups (labeled here as
# groups a, b, and c). By default, bitfields use 8-bit alignment, leading to the following
# layout:
#
#       Group      Pos       Bits
#       a          0x00      8
#       b          0x01      8
#       c          0x02      32
#
# Internally, only the first group requires special bit-level parsing. The remaining groups
# (b and c) are treated as standard structures since they span full bytes or words without
# sub-byte alignment. This dynamic grouping mechanism allows leveraging full struct-like
# class definitions within bitfields.
#
# This new approach enables more complex and expressive bitfield definitions. The annotation
# syntax is therefore extended as follows:
#
#   +---------------------------------------------------+--------------------------------------+
# 1.| <name> : <bits> [ - <field> ]                     | Standard field with optional type    |
#   +---------------------------------------------------+--------------------------------------+
# 2.| <name> : 0                                        | Aligns to the next byte boundary     |
#   +---------------------------------------------------+--------------------------------------+
# 3.| <name> : <field>                                  | Struct-like field (no bits consumed) |
#   +---------------------------------------------------+--------------------------------------+
# 4.| <name> : (<field>,<factory>)                      | Field with custom type factory       |
#   +---------------------------------------------------+--------------------------------------+
# 5.| <name> : (<bits>,<factory>[,<options>])           | bits with custom type factory        |
#   |        : (<bits>,[<options>])                     | and options                          |
#   +---------------------------------------------------+--------------------------------------+
#
# Processing Rules:
#
#   Rule 1.:
#       - Default alignment is 1 byte (8 bits).
#       - If followed by a rule 2 declaration, the remaining bits in the current byte are padded.
#       - If a <field> is provided:
#           - typeof(<field>) is used to infer the factory.
#           - etbits() and sizeof() determine the field’s alignment.
#       - If a custom alignment is configured in the Bitfield constructor, inferred alignment is
#         ignored unless the field includes the B_OVERWRITE_ALIGNMENT option.
#       - If the B_GROUP_END option is set, the current group is finalized and a new one is started.
#
#   Rule 2.:
#       - This rule forces alignment to the next byte boundary.
#       - The field is ignored during final class generation (name is discarded).
#       - The current group is finalized unless the bitfield is configured with B_GROUP_KEEP
#
#   Rule 3.:
#       - Equivalent to struct-like class field definitions.
#       - Automatically implies a rule 2 alignment.
#       - Always finalizes the current group regardless of B_GROUP_KEEP.
#
#   Rule 4.:
#       - Extension of Rule 1.
#       - Explicitly defines a conversion factory for the field.
#       - The factory must be:
#           - A built-in type (e.g., int, bool) supporting __int__, or
#           - A type or instance of BitfieldValueFactory.
#
#   Rule 5.:
#       - Builds upon Rule 4 with support for options.
#       - Options can be passed as a list or single element.
#       - Supported Options:
#           - NewGroup: Aligns the current group, starts a new one, and adds the entry to it.
#           - EndGroup: Adds the entry to the current group, then aligns it.
#           - SetAlignment: Changes the current working alignment.
#       - Note: Option order affects behavior and must be considered carefully.

#: The default alignment (in bits) used for bitfield group boundaries
DEFAULT_ALIGNMENT = 8

#: Alias for the `B_GROUP_NEW` flag, used to indicate that a new bitfield group should be started.
NewGroup = B_GROUP_NEW

#: Alias for the `B_GROUP_END` flag, used to indicate that the current bitfield group
#: should be finalized.
EndGroup = B_GROUP_END


class SetAlignment:
    """
    Instructional flag used to update the current bitfield alignment dynamically during
    bitfield generation.

    This class allows to explicitly set a new alignment boundary (in bits) for subsequent fields
    or groups in a bitfield definition. This enables finer control over how bitfield groups are
    organized and aligned.

    :param new_alignment: The alignment size in bits to be used from this point forward in the bitfield layout.
    :type new_alignment: int
    """

    def __init__(self, new_alignment: int) -> None:
        self.alignment = new_alignment

    @staticmethod
    def flag(new_alignment: int):
        """Create a :class:`Flag` instance representing a request to set a new alignment.

        This method is intended for use where a generic :class:`Flag` is expected rather than a full
        :class:`SetAlignment` object, e.g. for setting options for a :class:`Field`.

        >>> field = 5 - uint32 | SetAlignment.flag(32)

        :param new_alignment: The alignment size in bits.
        :type new_alignment: int
        :return: A `Flag` object with the key `"bitfield.new_alignment"` and the specified alignment as its value.
        :rtype: Flag
        """
        return Flag("bitfield.new_alignment", new_alignment)

    def __hash__(self) -> int:
        return hash("SetAlignment")


def getbits(obj) -> int:
    """Retrieve the bit-width of a given object.

    This function checks for a :py:func:`__bits__` attribute on the object. The object must either implement
    the :class:`_SupportsBits` or :class:`_ContainsBits` protocol.

    >>> class A:
    ...     __bits__ = 3
    ...
    >>> a = A()
    >>> getbits(a)
    3

    :param obj: The object for which the bit-width should be determined. It is expected to have an :attr:`ATTR_BITS` attribute.
    :type obj: Any
    :return: The number of bits used by the object.
    :rtype: int
    :raises AttributeError: If the object does not have an attribute defined by :attr:`ATTR_BITS`.
    """
    __bits__ = getattr(obj, ATTR_BITS)
    return __bits__() if callable(__bits__) else __bits__


def issigned(obj) -> bool:
    """Determine whether a given object represents a signed field.

    :param obj: The object for which signedness should be determined.
    :type obj: Any
    :return: :code:`True` if the field is marked as signed, :code:`False` otherwise.
    :rtype: bool
    """
    return bool(getattr(obj, ATTR_SIGNED, None))


class BitfieldValueFactory:
    """
    A generic factory class responsible for converting values between Python objects and integers
    for use in bitfield entries.

    By default, the factory converts to and from Python's built-in :code:`int` type, but it can be customized
    to support any type that accepts an integer in its constructor and implements :code:`__int__`.

    :param target:  The target type to which integer values will be converted., defaults to None
    :type target: type, optional
    """

    __slots__ = ("target",)

    def __init__(self, target=None) -> None:
        self.target = target or int

    def to_int(self, obj) -> int:
        """Convert a Python object to an integer.

        :param obj: The object to convert.
        :type obj: Any
        :return: The integer representation of the object.
        :rtype: int
        """
        return int(obj)

    def from_int(self, value: int):
        """Convert an integer to the target object type.

        :param value: The integer to convert.
        :type value: int
        :return: The value converted to the target type.
        :rtype: Any
        """
        return self.target(value)


class EnumFactory(BitfieldValueFactory):
    """A value factory for enum-like types used in bitfields.

    This factory attempts to convert between integers and enumeration instances,
    using the provided :code:`model` (which should support :code:`__int__`). It
    can operate in strict or lenient mode:

    - In strict mode, a :class:`ValueError` is raised if conversion fails.
    - In lenient mode, the raw integer is returned if the value is not in the enum.

    :param model: The enum model or mapping type to use.
    :type model: Type
    :param strict: Whether to raise an error on unknown values.
    :type strict: bool

    .. code-block:: python
        :caption: Example

        class Status(enum.IntEnum):
            OK = 0
            ERROR = 1

        factory = EnumFactory(Status, strict=True)
        factory.from_int(0)  # -> Status.OK
        factory.from_int(2)  # -> ValueError (strict mode)
    """

    def __init__(self, model, strict=False) -> None:
        super().__init__(model)
        self.strict = strict

    def from_int(self, value: int):
        """
        Convert an integer into an enum instance or raw int.

        :param value: The integer to convert.
        :type value: int
        :return: Enum instance or raw int if not found (in non-strict mode).
        :rtype: Any
        :raises ValueError: If strict is enabled and value is not valid.
        """
        try:
            return super().from_int(value)
        except ValueError:
            if self.strict:
                raise
            return value


class CharFactory(BitfieldValueFactory):
    """
    A value factory for handling single ASCII/Unicode characters as integers.

    This factory allows treating a character field as a one-byte integer and vice versa,
    automatically converting during packing and unpacking.
    """

    def __init__(self) -> None:
        super().__init__(str)

    def from_int(self, value: int):
        """
        Convert an integer to its character representation.

        :param value: Integer ASCII or Unicode code point.
        :type value: int
        :return: Corresponding character.
        :rtype: str
        """
        return chr(value)

    def to_int(self, obj) -> int:
        """
        Convert a character to its integer (ordinal) representation.

        :param obj: The character to convert.
        :type obj: str
        :return: Corresponding integer value.
        :rtype: int
        """
        return ord(obj)


class BitfieldEntry:
    """
    Represents a single entry in a bitfield, including its bit position, width, name, and
    conversion behavior.

    May also represent a special action or directive instead of a field.

    :param bit: The starting bit position within its group.
    :type bit: int
    :param width: The number of bits used by this field.
    :type width: int
    :param name: The name of the field.
    :type name: str
    :param factory: A factory for type conversion. Defaults to BitfieldValueFactory.
    :type factory: type or BitfieldValueFactory or None
    :param action: Optional action object for special handling (e.g., alignment or padding).
    :type action: Any
    """

    __slots__ = ("bit", "width", "name", "factory", "action", "low_mask")

    def __init__(
        self, bit: int, width: int, name: str, factory=None, action=None
    ) -> None:
        self.bit = bit
        self.width = width
        self.name = name
        self.factory = factory or BitfieldValueFactory()
        if isinstance(self.factory, type):
            self.factory = self.factory()
        self.action = action
        self.low_mask = (1 << self.width) - 1

    @staticmethod
    def new_action(action):
        """
        Create a new action-type entry (e.g., padding, control directive).

        :param action: The action object to encapsulate.
        :type action: Any
        :return: A BitfieldEntry instance with no bit-width, used for meta instructions.
        :rtype: BitfieldEntry
        """
        return BitfieldEntry(0, 0, "<action>", action=action)

    def shift(self, value_width: int) -> int:
        """
        Calculate how much to shift the field when extracting it from a value.

        :param value_width: The total bit width of the container.
        :type value_width: int
        :return: The number of bits to shift.
        :rtype: int
        """
        bit_pos = max(0, value_width - self.bit)
        return max(bit_pos - self.width, 0)

    def is_action(self) -> bool:
        """
        Check whether this entry is an action (i.e., not a data field).

        :return: True if this is an action entry.
        :rtype: bool
        """
        return self.action is not None

    def __repr__(self) -> str:
        """
        Return a human-readable string representation of the bitfield entry.

        :return: String representation.
        :rtype: str
        """
        if self.is_action():
            return repr(self.action)

        r = f"<Entry {self.name!r}: bit={self.bit}, width={self.width}"
        if self.factory:
            type_ = type(self.factory)
            if type_ is BitfieldValueFactory:
                type_ = self.factory.target

            r = f"{r}, factory={type_.__name__}"
        return f"{r}>"


class BitfieldGroup:
    """
    A group of one or more bitfield entries. Groups are used to organize fields within a single
    alignment unit and may represent either packed fields or standalone fields.

    :param bit_count: The number of bits in the group, or -1 for single field representation.
    :type bit_count: int
    """

    __slots__ = ("entries", "bit_count")

    def __init__(self, bit_count: int) -> None:
        self.bit_count = bit_count
        self.entries = []

    def is_field(self) -> bool:
        """
        Determine whether the group contains a single non-bitfield field.

        :return: True if the group holds a single struct-like field.
        :rtype: bool
        """
        return self.bit_count == -1

    def get_field(self):
        """
        Get the single field from this group.

        :return: The field object.
        :rtype: BitfieldEntry
        """
        return self.entries[0]

    def set_field(self, field):
        """
        Set the group to hold only the given field and mark it as a standalone field group.

        :param field: The field to store in this group.
        :type field: BitfieldEntry
        """
        self.entries = [field]
        self.bit_count = -1

    def align_to(self, alignment: int):
        """
        Align the bit count of this group to the specified boundary.

        :param alignment: The number of bits to align to.
        :type alignment: int
        """
        if not self.is_field():
            pad = self.bit_count % alignment
            if pad > 0:
                self.bit_count += alignment - pad

    def is_empty(self) -> bool:
        """
        Check if the group contains any entries.

        :return: True if the group is empty.
        :rtype: bool
        """
        return len(self.entries) == 0

    def get_size(self, context=None):
        """
        Get the size of this group in bytes.

        :param context: Optional context used for size evaluation.
        :type context: Any
        :return: The size of the group in bytes.
        :rtype: int
        """
        if self.is_field():
            field = self.get_field()
            return field.__size__(context) if context else sizeof(field)

        return self.bit_count // 8

    def get_bits(self, context=None):
        """
        Get the total number of bits in this group.

        :param context: Optional context used for size evaluation.
        :type context: Any
        :return: The number of bits.
        :rtype: int
        """
        return self.get_size(context) * 8

    def __repr__(self) -> str:
        if self.is_field():
            return repr(self.get_field())

        return f"<BitfieldGroup ({self.bit_count}) {self.entries!r}>"


class Bitfield(Struct):
    """
    A Bitfield represents a packed structure composed of bit-level fields. This
    class allows for the declarative definition of compact memory representations
    where each field can occupy an arbitrary number of bits, not necessarily
    aligned to byte boundaries.

    Core Implementation:
    - Bitfields are organized into BitfieldGroups, which manage alignment and field aggregation.
    - Entries can be individual bit widths or wrapped fields with explicit alignment.
    - Special field options like NewGroup and EndGroup can control group layout.
    - Supports value factories for type conversion and symbolic runtime actions.

    Available global options:
    - :code:`B_NO_AUTO_BOOL`: disables automatically converting 1bit fields to boolean
    - :code:`B_GROUP_KEEP`: disables finalizing groups when using the alignment definition syntax

    :param model: The model for the structure.
    :type model: Any
    :param order: Byte order of the structure.
    :type order: Optional[str]
    :param arch: Target architecture.
    :type arch: Optional[str]
    :param options: Global structure options.
    :type options: Optional[set]
    :param field_options: Field-specific options.
    :type field_options: Optional[set]
    :param alignment: Bit alignment size.
    :type alignment: Optional[int]
    """

    __slots__ = (
        "groups",
        "alignment",
    )

    def __init__(
        self,
        model,
        order=None,
        arch=None,
        options=None,
        field_options=None,
        alignment=None,
    ) -> None:
        self.alignment = alignment or DEFAULT_ALIGNMENT
        # These fields remain private and will be deleted after processing
        self._current_alignment = self.alignment
        self._current_group = BitfieldGroup(self._current_alignment)
        self._bit_pos = 0
        self.groups = [self._current_group]
        super().__init__(
            model=model,
            order=order,
            arch=arch,
            options=options,
            field_options=field_options,
        )
        # Add additional options based on the struct's type
        self.options.difference_update(GLOBAL_STRUCT_OPTIONS, GLOBAL_UNION_OPTIONS)
        self.options.update(GLOBAL_BITFIELD_FLAGS)

        self.groups = [group for group in self.groups if not group.is_empty()]
        self.groups[-1].align_to(self._current_alignment)
        # REVISIT: should be enable modification after processing?
        del self._bit_pos
        del self._current_alignment
        del self._current_group

    def __add__(self, sequence):
        """
        Append another Bitfield instance to this one.

        :param sequence: Another Bitfield instance.
        :type sequence: Bitfield
        :return: Combined Bitfield.
        :rtype: Bitfield
        :raises TypeError: If sequence is not a Bitfield.
        """
        if not isinstance(sequence, Bitfield):
            # REVISIT: we could just add each field as a group individually?
            raise TypeError(
                f"Attempted to add a non-bitfield struct to a bitfield! (type={type(sequence)})"
            )

        self.groups.extend(sequence.groups)
        return super(Struct, self).__add__(sequence)

    def _process_align(self, options) -> Field:
        """
        Process an alignment directive.

        .. code-block:: bnf

            <name> : 0

        :param options: A list of alignment-related options.
        :type options: Optional[list]
        :return: A placeholder field.
        :rtype: Field
        """
        # 2.: the current group will be finalized
        self._current_group.align_to(self._current_alignment)
        if not self.has_option(B_GROUP_KEEP):
            self._current_group = self._new_group(self._current_alignment)
        else:
            self._bit_pos = self._current_group.bit_count

        for option in options or []:
            if self._process_alignment_option(option):
                continue

            alignment = self._current_alignment
            group = self._current_group
            if option.name in (EndGroup.name, NewGroup.name):
                # finalize current group (same effect for alignment statement)
                group.align_to(alignment)
                self._current_group = self._new_group(alignment)

        return Field(Pass)

    def _process_bits(self, name: str, bits: int, factory=None, options=None) -> Field:
        """
        Process a bitfield entry with a given width.

        :param name: Field name.
        :type name: str
        :param bits: Width in bits.
        :type bits: int
        :param factory: Optional value factory.
        :type factory: Optional[BitfieldValueFactory]
        :param options: Field-specific options.
        :type options: Optional[list]
        :return: Resulting Field.
        :rtype: Field
        """
        if bits == 0:
            return self._process_align(options)

        if not factory and bits == 1 and not self.has_option(B_NO_AUTO_BOOL):
            factory = BitfieldValueFactory(bool)

        entry = BitfieldEntry(self._bit_pos, bits, name, factory)
        if not self._process_options(options, entry):
            group = self._current_group
            group.entries.append(entry)
            # Adjust the size of the goup dynamically
            self._bit_pos += entry.width
            group.bit_count = max(group.bit_count, self._bit_pos)
        # this is only symbolic
        return Field(Int(bits))

    def _process_bits_field(
        self, name: str, field, options=None, factory=None
    ) -> Field:
        """
        Process a bitfield that wraps another field instance.

        :param name: Field name.
        :type name: str
        :param field: The field instance.
        :type field: Field
        :param options: List of options.
        :type options: Optional[list]
        :param factory: Optional value factory.
        :type factory: Optional[BitfieldValueFactory]
        :return: Resulting Field.
        :rtype: Field
        :raises TypeError: If bit width is not an integer.
        """
        if field.bits is None:
            # we don't need to check for NewGroup and EndGroup options here as no
            # bits are specified and the field gets its own group.
            for option in options or []:
                self._process_alignment_option(option)

            # bits not present -> treat defintion as simple field, which means we finalize
            # the current group, create a new FIELD GROUP and another new one after that
            self._current_group.align_to(self._current_alignment)
            field_group = self._new_group(-1)
            field_group.set_field(field)
            self._current_group = self._new_group(self._current_alignment)
            return field

        width = field.bits
        if not isinstance(width, int):
            raise TypeError(
                f"Bitfield: field definition of {name!r} requires an integer as bits"
            )

        entry = BitfieldEntry(
            self._bit_pos, width, name, factory or BitfieldValueFactory(typeof(field))
        )
        if not self._process_options(options, entry):
            group = self._current_group
            group.entries.append(entry)
            self._bit_pos += width
            group.bit_count = max(group.bit_count, self._bit_pos)

        if field.has_flag(B_OVERWRITE_ALIGNMENT):
            try:
                field_bits = getbits(field.struct)
            except Exception:
                field_bits = sizeof(field) * 8

            self._current_alignment = field_bits or DEFAULT_ALIGNMENT
            self._current_group.align_to(self._current_alignment)

        return field

    def _process_options(self, options, entry=None) -> bool:
        consumed = False
        for option in options or []:
            if self._process_alignment_option(option):
                continue

            group = self._current_group
            alignment = self._current_alignment
            if option.name == EndGroup.name:
                if entry:
                    group.entries.append(entry)
                    self._bit_pos += entry.width
                    group.bit_count = max(group.bit_count, self._bit_pos)
                    consumed = True

                group.align_to(alignment)
                self._current_group = self._new_group(alignment)
            elif option.name == NewGroup.name:
                # finalize current group, create a new one and add the entry to the newly
                # created group
                group.align_to(alignment)
                self._current_group = group = self._new_group(alignment)

                if entry:
                    group.entries.append(entry)
                    # position was reset to zero
                    self._bit_pos += entry.width
                    group.bit_count = max(group.bit_count, self._bit_pos)
                    consumed = True
        return consumed

    def _new_group(self, alignment):
        new_group = BitfieldGroup(alignment)
        self.groups.append(new_group)
        self._bit_pos = 0
        return new_group

    def _process_alignment_option(self, option):
        if isinstance(option, SetAlignment):
            # update current working alignment
            self._current_alignment = option.alignment or DEFAULT_ALIGNMENT
            self._current_group.align_to(self._current_alignment)
            return True
        elif isinstance(option, Flag):
            if option.name == "bitfield.new_alignment":
                self._current_alignment = option.value or DEFAULT_ALIGNMENT
                self._current_group.align_to(self._current_alignment)
                return True

        return False

    def _process_field(self, name: str, annotation, default):
        arch = self.arch or system_arch
        order = getattr(annotation, ATTR_BYTEORDER, self.order or SysNative)
        match annotation:
            case int():
                if annotation == 0:
                    return self._process_align(None)
                else:
                    # 1. (without field) defines the width using the default value factory
                    return self._process_bits(name, annotation)

            case tuple():  # NEW EXTENDED DEFINITION
                if len(annotation) == 0:
                    raise ValueError(
                        f"Extended field definition for {name!r} does not define any values!"
                    )

                if len(annotation) == 1:
                    # definition: (<bits>, )
                    (width,) = annotation
                    if width == 0:
                        raise ValueError(
                            f"Extended field {name!r} defines forbidden width of zero"
                        )

                    return self._process_bits(name, width)

                options = []
                factory = None
                (width, factory_or_option, *extra_options) = annotation
                if isinstance(factory_or_option, BitfieldValueFactory):
                    factory = factory_or_option
                elif isinstance(factory_or_option, type):
                    # here we enable just specifying types instead of a factory
                    if not issubclass(factory_or_option, BitfieldValueFactory):
                        factory = BitfieldValueFactory(factory_or_option)
                    else:
                        factory = factory_or_option()
                else:
                    # treat as option or as a list of options
                    options = (
                        [factory_or_option]
                        if not isinstance(factory_or_option, Iterable)
                        else list(factory_or_option)
                    )

                # extra options may be a list or single element
                for extra in extra_options:
                    options.extend(extra if isinstance(extra, Iterable) else [extra])

                if isinstance(width, int):
                    # rule no. 5
                    return self._process_bits(name, width, factory, options)

                field = width
                if not isinstance(field, Field):
                    field = Field(field, order=order, arch=arch, default=default)

                return self._process_bits_field(name, field, options, factory)

            case _:
                # rule 1 (with field) or rule 3
                field = self._process_annotation(annotation, default, order, arch)
                if not isinstance(field, Field):
                    field = Field(field, order=order, arch=arch, default=default)
                return self._process_bits_field(name, field)

    def _included(self, name: str, default, annotation) -> bool:
        if not super()._included(name, default, annotation):
            return False

        width = annotation
        if isinstance(annotation, tuple):
            width, *_ = annotation

        if isinstance(width, int):
            return width != 0

        if isinstance(width, Field):
            return width.bits != 0
        return True

    def _replace_type(self, name: str, type_: type) -> None:
        entry = self.get_entry(name)
        if entry is not None:
            if not entry.factory:
                type_ = int
            elif isinstance(entry.factory, BitfieldValueFactory):
                type_ = entry.factory.target or object
            else:
                type_ = object

        # else: must be a field with a known type
        return super()._replace_type(name, type_)

    def __size__(self, context) -> int:
        """
        Calculate the total size of the bitfield structure.

        :param context: Packing context.
        :type context: Any
        :return: Total size in bytes.
        :rtype: int
        """
        # size is different as our model includes correct padding
        return sum(map(lambda g: g.get_size(context), self.groups))

    def __bits__(self) -> int:
        """
        Compute the total number of bits in the structure.

        :return: Total bit count.
        :rtype: int
        """
        return sum(map(lambda g: g.get_bits(), self.groups))

    def unpack_one(self, context):
        init_data = O_CONTEXT_FACTORY.value()
        context[CTX_OBJECT] = O_CONTEXT_FACTORY.value(_parent=context)
        base_path = context[CTX_PATH]
        # REVISIT
        endian = "little" if self.order == LittleEndian else "big"
        for group in self.groups:
            if group.is_field():
                # unpack using field instance
                field = group.get_field()
                name = field.__name__
                context[CTX_PATH] = f"{base_path}.{name}"
                value = field.__unpack__(context)
                context[CTX_OBJECT][name] = value
                if name in self._members:
                    init_data[name] = value

            else:
                raw_data = context[CTX_STREAM].read(group.get_size())
                if not raw_data:
                    # set context path to next entry for debugging
                    context[CTX_PATH] = f"{base_path}.{group.entries[0].name}"
                    raise StructException(
                        f"Failed to parse group of size {group.bit_count}bits: unexpected EOF!",
                        context,
                    )
                raw_value = int.from_bytes(raw_data, endian)
                for entry in group.entries:
                    # each entry may be an action
                    context[CTX_PATH] = f"{base_path}.{entry.name}"
                    if entry.is_action():
                        func = getattr(entry.action, ATTR_ACTION_UNPACK, None)
                        if func:
                            func(context)
                            continue

                    if entry.name not in self._members:
                        continue

                    value = (raw_value >> entry.shift(group.bit_count)) & entry.low_mask
                    if entry.factory:
                        value = entry.factory.from_int(value)

                    init_data[entry.name] = value

        return self.model(**init_data)

    def pack_one(self, obj, context) -> None:
        base_path = context[CTX_PATH]
        # REVISIT
        endian = "little" if self.order == LittleEndian else "big"
        for group in self.groups:
            if group.is_field():
                field = group.get_field()
                name = field.__name__
                context[CTX_PATH] = f"{base_path}.{name}"
                if name in self._members:
                    value = self.get_value(obj, name, field)
                else:
                    value = field.default if field.default != INVALID_DEFAULT else None

                field.__pack__(value, context)
            else:
                value = 0
                for entry in group.entries:
                    context[CTX_PATH] = f"{base_path}.{entry.name}"
                    if entry.is_action():
                        func = getattr(entry.action, ATTR_ACTION_PACK, None)
                        if func:
                            func(context)
                        continue

                    if entry.name not in self._members:
                        continue

                    entry_value = self.get_value(obj, entry.name, None)
                    if entry.factory:
                        entry_value = entry.factory.to_int(entry_value)

                    # silently ignore invalid values
                    value |= (entry_value & entry.low_mask) << entry.shift(
                        group.bit_count
                    )
                context[CTX_STREAM].write(value.to_bytes(group.bit_count // 8, endian))

    def add_action(self, action) -> None:
        self._current_group.entries.append(
            BitfieldEntry(0, 0, "<action>", action=action)
        )
        return super().add_action(action)

    def get_entry(self, name: str):
        for group in self.groups:
            if group.is_field():
                continue

            for entry in group.entries:
                if entry.name == name:
                    return entry


def _make_bitfield(
    cls: type,
    /,
    *,
    options,
    order=None,
    arch=None,
    field_options=None,
    alignment=None,
) -> type:
    _ = Bitfield(
        cls,
        order=order,
        arch=arch,
        options=options,
        field_options=field_options,
        alignment=alignment,
    )
    return cls


def bitfield(
    cls=None,
    /,
    *,
    options=None,
    order=None,
    arch=None,
    field_options=None,
    alignment=None,
):
    """
    Decorator that transforms a class definition into a :class:`Bitfield` structure.

    This decorator enables defining bitfields using simple class syntax,
    with support for custom alignment, ordering, architecture, and field options.

    :param cls: The user-defined class to transform.
    :type cls: Optional[type]
    :param options: A set of global or structure-specific options.
    :type options: Optional[set]
    :param order: Optional byte order for serialization (e.g., 'little' or 'big').
    :type order: Optional[str]
    :param arch: Optional architecture string (e.g., 'x86', 'arm').
    :type arch: Optional[str]
    :param field_options: Optional default options for fields.
    :type field_options: Optional[set]
    :param alignment: Optional alignment in bits.
    :type alignment: Optional[int]
    :return: The decorated class, enhanced as a `Bitfield` structure.
    :rtype: type

    .. code-block:: python

        from caterpillar.py import bitfield, SetAlignment, uint16

        @bitfield
        class Packet:
            version   : 3
            type      : (5, SetAlignment(16))
            length    : 10
            _         : 0  # align to 16bits
            payload   : uint16

        # You can now pack/unpack Packet instances as compact binary bitfields
        pkt = Packet(version=1, type=2, length=128, payload=0xABCD)
        packed = pack(pkt)
        unpacked = unpack(Packet, packed)
    """

    def wrap(cls):
        return _make_bitfield(
            cls,
            options=options,
            order=order,
            arch=arch,
            field_options=field_options,
            alignment=alignment,
        )

    if cls is not None:
        return _make_bitfield(
            cls,
            options=options,
            order=order,
            arch=arch,
            field_options=field_options,
            alignment=alignment,
        )

    return wrap



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from sys import maxsize
from platform import machine
from dataclasses import dataclass
from enum import Enum

from caterpillar.shared import ATTR_BYTEORDER


@dataclass(frozen=True)
class ByteOrder:
    """
    Represents byte order information, including alignment and size.

    :param name: A string representing the name of the byte order.
    :param ch: A string representing the character used to specify the byte
               order in struct format strings.
    :param alignment: An enumeration representing the alignment (default
                      is `Alignment.NONE`).
    :param size: An enumeration representing the size (default is
                 `Size.STANDARD`).
    """

    name: str
    ch: str

    class Alignment(Enum):
        NONE = 0
        NATIVE = 1

    alignment: Alignment = Alignment.NONE

    class Size(Enum):
        STANDARD = 0
        NATIVE = 1

    size: Size = Size.STANDARD

    def apply(self, other):
        """
        Applies the byte order information to another object.

        :param other: The object to which the byte order information should be applied.
        """
        setattr(other, ATTR_BYTEORDER, self)

    def __add__(self, other):
        """
        Adds the byte order information to another object using the
        `__set_byteorder__` method.

        :param other: The object to which the byte order information should be added.
        :return: The modified object.
        """
        return other.__set_byteorder__(self)

    def __or__(self, other):
        """
        Applies the byte order information to another object using the `apply` method.

        :param other: The object to which the byte order information should be applied.
        :return: The modified object.
        """
        self.apply(other)
        return other


LITTLE_ENDIAN_FMT = "<"

# Instances representing commonly used byte orders
Native = ByteOrder("Native", "=")
BigEndian = ByteOrder("Big Endian", ">")
LittleEndian = ByteOrder("Little Endian", LITTLE_ENDIAN_FMT)
NetEndian = ByteOrder("Network", "!")
SysNative = ByteOrder(
    "SysNative", "@", ByteOrder.Alignment.NATIVE, ByteOrder.Size.NATIVE
)




def byteorder(obj, default=None):
    """
    Get the byte order of an object, defaulting to SysNative if not explicitly set.

    :param obj: The object to retrieve the byte order from.
    :return: The byte order of the object.
    """
    return getattr(obj, ATTR_BYTEORDER, default or SysNative)


@dataclass(frozen=True)
class Arch:
    """
    Represents a system architecture with a name and an indication of whether it is 64-bit.

    :param name: The name of the architecture.
    :param ptr_size: the amount of bits one pointer takes
    """

    name: str
    ptr_size: int


system_arch: Arch = Arch(machine(), 64 if maxsize > 2**32 else 32)

# common architectures
x86 = Arch("x86", 32)
x86_64 = Arch("x86-64", 64)
ARM = Arch("ARM", 32)
ARM64 = Arch("ARM64", 64)
AARCH64 = ARM64
PowerPC = Arch("PowerPC", 32)
PowerPC64 = Arch("PowerPC64", 64)
MIPS = Arch("MIPS", 32)
MIPS64 = Arch("MIPS64", 64)
SPARC = Arch("SPARC", 32)
SPARC64 = Arch("SPARC64", 64)
RISC_V64 = Arch("RISK-V64", 64)
RISC_V = Arch("RISK-V", 32)
AMD = Arch("AMD", 32)
AMD64 = Arch("AMD64", 64)



from caterpillar.shortcuts import struct, LittleEndian, bitfield, unpack, pack
from caterpillar.fields import uint8, uint24, CString, Prefixed, uint32


# The __slots__ options does not affect the packing or
# unpacking process.
# opt.set_struct_flags(opt.S_SLOTS, opt.S_REPLACE_TYPES)

# The following lines will reduce the time around 10% for
# packing and unpacking. (C-Extension required)
# from caterpillar.context import O_CONTEXT_FACTORY
# from caterpillar.c import c_Context
# O_CONTEXT_FACTORY.value = c_Context

@bitfield(order=LittleEndian)
class Flags:
    bool1: 1
    num4: 3
    # padding is generated automatically


@struct(order=LittleEndian)
class Item:
    num1: uint8
    num2: uint24
    flags: Flags
    fixedarray1: uint8[3]
    name1: CString()
    # Replacing the encoding with String(...) makes a huge difference
    # as we don't call .decode() directly
    #
    # Time goes down from 0.0119 to 0.0099 for unpacking
    # and from 0.0094 to 0.0082 for packing
    name2: Prefixed(uint8, encoding="utf-8")


# makes no real difference
# @struct(order=LittleEndian)
# class Format:
#     items: Item[uint32::]

Format = LittleEndian + Item[uint32::]

if __name__ == "__main__":
    import sys
    import timeit

    try:
        from rich import print
    except ImportError:
        pass

    with open(sys.argv[1], "rb") as fp:
        data = fp.read()

    obj = unpack(Format, data)
    time = timeit.timeit(lambda: unpack(Format, data), number=1000) / 1000
    print("[bold]Timeit measurements:[/]")
    print(f"[bold]unpack[/] {time:.10f} sec/call")

    ptime = timeit.timeit(lambda: pack(obj, Format), number=1000) / 1000
    print(f"[bold]pack[/]   {ptime:.10f} sec/call")



"""
Custom script to generate the public Caterpillar API.
=====================================================
"""

import pathlib
import argparse
import dataclasses
from typing_extensions import override

from caterpillar_api import (
    CP_FUNC_API,
    CP_TYPE_API,
    cp_api_functions,
    CP_SRC,
    CP_TYPES,
)


# Base class for all API objects (functions or types)
@dataclasses.dataclass
class APIObj:
    name: str
    index: int
    api_name: str
    type: str

    def cp_api_table(self) -> str:
        if not self.name:
            return "NULL"
        return f"(void *) &{self.name}"

    def cp_internal_def(self) -> str:
        # must be overridden in subclasses
        return ""

    def cp_external_def(self) -> str:
        return ""


class APIType(APIObj):
    def __insit__(
        self, name: str, index: int, api_name: str, type_: str | None = None
    ) -> None:
        super().__init__(name, index, api_name, type_ or "PyTypeObject")

    @override
    def cp_internal_def(self) -> str:
        # REVISIT: what about struct definitions?
        if self.type.startswith("Py"):
            return f"extern {self.type} {self.name};"

        extra_def = f"#define {self.name.replace('_Type', '')}_NAME \"{self.type}\""
        return f"extern PyTypeObject {self.name};\n{extra_def}"

    @override
    def cp_external_def(self):
        """Return the external definition for this API object

        Example: Consider the simple C type `CpAtom_Type`. Its external definition
        will be `#define CpAtom_Type (*(PyTypeObject *)Cp_API[0])` in case `CpAtom`
        is assigned to index zero.

        :return: C source code string
        :rtype: str
        """
        return f"#define {self.name} (*({self.type} *){self.api_name}[{self.index}])"


class APIFunc(APIObj):
    rtype: str
    args: list[str]

    def __init__(
        self, name: str, index: int, api_name: str, rtype: str, args: list[str]
    ) -> None:
        self.rtype = rtype
        self.args = args
        type_ = "(%s (*)(%s))" % (rtype, ", ".join(args))
        super().__init__(name, index, api_name, type_)

    @override
    def cp_internal_def(self) -> str:
        return f"{self.rtype} {self.name}({', '.join(self.args)});"

    @override
    def cp_external_def(self) -> str:
        return f"#define {self.name} (*({self.type}){self.api_name}[{self.index}])"


def genapi(
    api_h: pathlib.Path,
    api_c: pathlib.Path,
    mod_c: pathlib.Path,
    private_h: pathlib.Path,
) -> None:
    api_h_out = pathlib.Path(str(api_h).replace(".in", ""))
    api_c_out = pathlib.Path(str(api_c).replace(".in", ""))
    mod_c_out = pathlib.Path(str(mod_c).replace(".in", ""))
    private_h_out = pathlib.Path(str(private_h).replace(".in", ""))

    func_api = cp_api_functions()
    max_func_index = max(CP_FUNC_API.values())
    max_type_index = max(map(lambda x: x[0], CP_TYPE_API.values()))

    api_objects = [None] * (max(max_func_index, max_type_index) + 1)
    for name, args in CP_TYPE_API.items():
        if name == "__reserved__":
            continue
        index = args[0]
        if api_objects[index] is not None:
            raise ValueError(f"Duplicate API object: {name} at index {index}")

        type_ = None if len(args) == 1 else args[1]
        api_objects[index] = APIType(name, index, "Cp_API", type_)

    for name, index in CP_FUNC_API.items():
        if api_objects[index] is not None:
            raise ValueError(f"Duplicate API object: {name} at index {index}")

        rtype, args = func_api[name]
        api_objects[index] = APIFunc(name, index, "Cp_API", rtype, args)

    array_def = []
    external_def = []
    internal_def = []
    for obj in api_objects:
        if not obj:
            array_def.append("NULL")
        else:
            array_def.append(obj.cp_api_table())
            external_def.append(obj.cp_external_def())
            internal_def.append(obj.cp_internal_def())

    typedefs = []
    for struct_name, type_name in CP_TYPES.items():
        typedefs.append(f"struct {struct_name};")
        typedefs.append(f"typedef struct {struct_name} {type_name};")

    api_h_source = api_h.read_text(encoding="utf-8")
    mod_types = []
    for name, ignored in CP_SRC:
        if not ignored:
            target_name = pathlib.Path(name).stem
            mod_types.append(f"cp_{target_name}")

    with api_h_out.open("w", encoding="utf-8") as f:
        f.write(
            api_h_source
            % ("\n".join(typedefs), "\n".join(internal_def), "\n".join(external_def))
        )

    api_c_source = api_c.read_text(encoding="utf-8")
    with api_c_out.open("w", encoding="utf-8") as f:
        f.write(api_c_source % (",\n    ".join(array_def),))

    mod_source = mod_c.read_text(encoding="utf-8")
    mod_c_out.write_text(
        mod_source
        % (
            "\n    ".join([f"{name}__mod_clear(m, state);" for name in mod_types]),
            "\n  ".join([f"SETUP_TYPES({name})" for name in mod_types]),
            "\n  ".join([f"ADD_OBJECTS({name})" for name in mod_types]),
        ),
        encoding="utf-8",
    )

    private_h_source = private_h.read_text(encoding="utf-8")
    private_h_out.write_text(
        private_h_source % ("\n".join([f"_CpDef_ModFn({name});" for name in mod_types]))
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("api_header", type=pathlib.Path)
    parser.add_argument("api_source", type=pathlib.Path)
    parser.add_argument("mod_source", type=pathlib.Path)
    parser.add_argument("private_header", type=pathlib.Path)

    argv = parser.parse_args()
    genapi(argv.api_header, argv.api_source, argv.mod_source, argv.private_header)



from ._mixin import FieldStruct
from caterpillar.abc import _ContextLike
from caterpillar.options import Flag
from typing import Tuple, Type

VARINT_LSB: Flag[None]

class VarInt(FieldStruct[int, int]):
    def __type__(self) -> Type[int]: ...
    def __size__(self, context: _ContextLike) -> int: ...
    def bit_config(self, context: _ContextLike) -> Tuple[int, int]: ...
    def pack_single(self, obj: int, context: _ContextLike) -> None: ...
    def unpack_single(self, context: _ContextLike) -> int: ...

vint: VarInt



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from ._base import Sequence
from ._struct import (
    Struct,
    struct,
    UnionHook,
    union,
    unpack,
    unpack_file,
    pack,
    pack_into,
    pack_file,
    sizeof,
)
from ._bitfield import (
    Bitfield,
    bitfield,
    BitfieldEntry,
    BitfieldGroup,
    BitfieldValueFactory,
    issigned,
    getbits,
    NewGroup,
    EndGroup,
    SetAlignment,
    EnumFactory,
    CharFactory,
    DEFAULT_ALIGNMENT,
)
from ._template import istemplate, template, TemplateTypeVar, derive

__all__ = [
    "Sequence",
    "Struct",
    "struct",
    "UnionHook",
    "union",
    "unpack",
    "unpack_file",
    "pack",
    "pack_into",
    "pack_file",
    "sizeof",
    "Bitfield",
    "bitfield",
    "BitfieldGroup",
    "issigned",
    "getbits",
    "istemplate",
    "template",
    "TemplateTypeVar",
    "derive",
    "NewGroup",
    "EndGroup",
    "SetAlignment",
    "BitfieldEntry",
    "BitfieldValueFactory",
    "EnumFactory",
    "CharFactory",
    "DEFAULT_ALIGNMENT",
]



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
try:
    from rich import print
except ImportError:
    pass

# --------------------------- Basic Information ---------------------------
# "Apple’s Core Audio Format (CAF) is a file format for storing and
# transporting digital audio data. It simplifies the management and
# manipulation of many types of audio data without the file-size
# limitations of other audio file formats."
# > https://developer.apple.com/library/archive/documentation/MusicAudio/Reference/CAFSpec/CAF_intro/CAF_intro.html
# -------------------------------------------------------------------------
import enum

from caterpillar.shortcuts import BigEndian, this, parent, opt
from caterpillar.model import struct, unpack_file, pack_file
from caterpillar.fields import *


opt.set_struct_flags(opt.S_SLOTS)
try:
    from numpy import array

    opt.O_ARRAY_FACTORY.value = array
except ImportError:
    pass


@struct(order=BigEndian)
class CAFHeader:
    # These fields follow standard the definition without a direct
    # default value
    file_type: b"caff"
    file_version: uint16
    # according to Apple, this field must be 0 for CAF c1 files and is reserved
    # for future usage
    file_flags: uint16


@struct(order=BigEndian)
class CAFChunkHeader:
    # As we don't know all possible chunk types, it is recommended to
    # use the Bytes struct here instead of an enum class.
    chunk_type: Bytes(4)
    chunk_size: uint64


# Enum classes are defined as simple as this. Note that the specified enum value
# should be returned by the parsing struct.
class AudioFormatFlags(enum.IntEnum):
    CAFLinearPCMFormatFlagIsFloat = 1
    CAFLinearPCMFormatFlagIsLittleEndian = 0


@struct(order=BigEndian)
class CAFAudioFormat:
    sample_rate: float64
    # The explicit encoding here is for illustration purposes. The default
    # one is actually utf-8.
    format_id: String(4, encoding="utf-8")
    format_flags: Enum(AudioFormatFlags, uint32)
    # The next fields are all simple integer fields
    bytes_per_packet: uint32
    frames_per_packet: uint32
    channels_per_frame: uint32
    bits_per_channel: uint32


@struct(order=BigEndian)
class CAFData:
    edit_count: uint32
    # The parent object can be referenced with a simple shortcut that accesses
    # the direct parent object context.
    data: Memory(parent.chunk_header.chunk_size - 4)


@struct(order=BigEndian)
class CAFInformation:
    # Here we use the _GreedyType to indicate that we don't known the size of
    # the string to parse. The struct will parse until it reads the termination
    # character '\x00'
    key: CString(...)
    value: CString(...)


@struct(order=BigEndian)
class CAFStringsChunk:
    # A simple prefixed field where we use a reference to an already parsed value
    # as the length.
    # num_entries: uint32
    strings: CAFInformation[uint32::]


@struct(order=BigEndian)
class CAFPacketTable:
    # note all signed integer fields here
    num_packets: int64
    num_valid_frames: int64
    priming_frames: int32
    remainder_frames: int32
    # The VarInt configuration doesn't have to be changed, as this file format uses
    # the default implementation.
    table_data: vint[this.num_packets]


@struct(order=BigEndian)
class CAFChannelDesc:
    channel_label: uint32
    channel_flags: uint32
    # fixed size arrays are also possible
    coordinates: float32[3]


@struct(order=BigEndian)
class CAFChannelLayout:
    channel_layout_tag: uint32
    channel_bitmap: uint32
    # REVISIT: the field 'num_descriptions' is obsolete here. There should be a
    # possibility to incorporate the field automatically into arrays.
    num_descriptions: uint32
    descriptions: CAFChannelDesc[this.num_descriptions]


@struct(order=BigEndian)
class CAFChunk:
    chunk_header: CAFChunkHeader
    # this is another simple switch-case structure, but this time we define a
    # default option. NOTE: not all chunk types are implemented
    data: Field(this.chunk_header.chunk_type) >> {
        b"desc": CAFAudioFormat,
        b"info": CAFStringsChunk,
        b"pakt": CAFPacketTable,
        b"data": CAFData,
        # In order to save memory space, we can define a simple padding that inserts
        # the missing data automatically
        b"free": padding[this.chunk_header.chunk_size],
        # NOTE: we use the special struct 'Memory' here as it makes the output more
        # clear (bytes would consume unnecessary decoding time)
        DEFAULT_OPTION: Memory(this.chunk_header.chunk_size),
    }


@struct(order=BigEndian)
class CAF:
    # interestingly, this file format does not declare how many chunks are
    # stored in one file, so we have to use greedy parsing here.
    header: CAFHeader
    chunks: CAFChunk[...]


if __name__ == "__main__":
    import sys

    obj = unpack_file(CAF, sys.argv[1])
    print(obj)
    pack_file(obj, sys.argv[2])



from rich import print
from cryptography.hazmat.primitives import hashes

from caterpillar.py import struct, Bytes, pack, unpack, this
from caterpillar.fields.digest import HMAC


@struct
class Format:
    key: b"MAGIC"

    with HMAC(this.key, hashes.SHA256(), "hmac", verify=True):
        user_data: Bytes(11)

    # attribute 'hmac' is added by HMAC


print(Format(user_data=b"hello world"))
print(Format.__struct__.fields)
data = pack(Format(user_data=b"hello world"))
print(data)

print("Unpacking invalid data:")
print(unpack(Format, data[:-1] + b"\x00"))



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from caterpillar import registry
from caterpillar.abc import (
    _ContextLambda,
    _ContextLike,
    _GreedyType,
    _PrefixedType,
    _StructLike,
    _SwitchLike,
    _IT,
    _OT,
    _LengthT,
    _StructT,
    _OptionLike,
    _EndianLike,
)
from caterpillar.byteorder import (
    Arch,
    ByteOrder,
)
from typing import Any, TypeVar
from typing_extensions import Self

_T = TypeVar("_T")

def singleton(cls: type[_T]) -> _T: ...

INVALID_DEFAULT: object = ...
DEFAULT_OPTION: object = ...

class Field(_StructLike[_IT, _OT]):
    order: _EndianLike
    arch: Arch
    default: _OT | None
    bits: _ContextLambda | int | None
    __name__: str | None
    def __init__(
        self,
        struct: _StructT[_IT, _OT],
        order: _EndianLike | None = None,
        offset: _ContextLambda | int = -1,
        flags: set[_OptionLike] | None = None,
        amount: _ContextLambda | int | _PrefixedType = 0,
        options: _SwitchLike[_OT, Any] | None = None,
        condition: _ContextLambda | bool = True,
        arch: Arch | None = None,
        default: _OT | None = ...,
        bits: _ContextLambda | int | None = None,
    ) -> None: ...
    @property
    def struct(self) -> _StructT[_IT, _OT]: ...
    @struct.setter
    def struct(self, value: _StructT[_IT, _OT]) -> None: ...
    @property
    def condition(self) -> _ContextLambda[bool] | bool: ...
    @condition.setter
    def condition(self, value: _ContextLambda[bool] | bool) -> None: ...
    @property
    def flags(self) -> set[_OptionLike]: ...
    @flags.setter
    def flags(self, value: set[ _OptionLike]) -> None: ...
    def add_flag(self, flag: _OptionLike) -> None: ...
    def has_flag(self, flag: _OptionLike) -> bool: ...
    def remove_flag(self, flag: _OptionLike) -> None: ...
    @property
    def offset(self) -> _ContextLambda[int] | int: ...
    @offset.setter
    def offset(self, value: _ContextLambda[int] | int) -> None: ...
    @property
    def amount(self) -> _LengthT: ...
    @amount.setter
    def amount(self, value: _LengthT) -> None: ...
    @property
    def options(self) -> _SwitchLike[_OT, Any]: ...
    @options.setter
    def options(self, value: _SwitchLike[_OT, Any]) -> None: ...
    def __or__(self, flag: _OptionLike) -> Self: ...
    def __xor__(self, flag: _OptionLike) -> Self: ...
    def __matmul__(self, offset: _ContextLambda | int) -> Self: ...
    def __getitem__(self, dim: _LengthT) -> Self: ...
    def __rshift__(self, switch: _SwitchLike[_IT, _OT]) -> Self: ...
    def __floordiv__(self, condition: _ContextLambda | bool) -> Self: ...
    def __rsub__(self, bits: _ContextLambda | int) -> Self: ...
    def __set_byteorder__(self, order: ByteOrder) -> Self: ...
    def __type__(self) -> type: ...
    def __unpack__(self, context: _ContextLike) -> _OT: ...
    def __pack__(self, obj: _IT, context: _ContextLike) -> None: ...
    def __size__(self, context: _ContextLike) -> int: ...
    __ixor__ = __xor__
    __ior__ = __or__
    __ifloordiv__ = __floordiv__
    __irshift__ = __rshift__
    __imatmul__ = __matmul__
    __isub__ = __rsub__
    def _verify_context_value(self, value: Any, expected: type) -> None: ...
    def is_seq(self) -> bool: ...
    def is_enabled(self, context: _ContextLike) -> bool: ...
    def has_condition(self) -> bool: ...
    def length(self, context: _ContextLike) -> int | _GreedyType | _PrefixedType: ...
    def get_struct(
        self, value: Any, context: _ContextLike
    ) -> _StructLike[_IT, _OT]: ...
    def get_offset(self, context: _ContextLike) -> int: ...
    def get_type(self) -> type: ...
    def get_name(self) -> str | None: ...

class _CallableTypeConverter(registry.TypeConverter):
    def matches(self, annotation: Any) -> bool: ...
    def convert(self, annotation: Any, kwargs: dict) -> Field: ...



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from ._base import Field
from ._mixin import FieldStruct
from _typeshed import SupportsLenAndGetItem

from caterpillar import registry
from caterpillar.abc import (
    _ContextLambda,
    _ContextLike,
    _LengthT,
    _StructLike,
    _IT,
    _OT,
)
from caterpillar.options import Flag
from functools import cached_property
from types import EllipsisType, NoneType
from typing import (
    Any,
    Callable,
    Collection,
    Final,
    Generic,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
    override,
    type_check_only,
)
from uuid import UUID

ENUM_STRICT: Flag[NoneType] = ...

class PyStructFormattedField(FieldStruct[_IT, _IT]):
    text: str
    ty: Type[_IT]
    __bits__: int
    def __init__(self, ch: str, type_: Type[_IT]) -> None: ...
    def __type__(self) -> type: ...
    def __size__(self, context: _ContextLike) -> int: ...
    def pack_single(self, obj: _IT, context: _ContextLike) -> None: ...
    def pack_seq(self, seq: Collection[_IT], context: _ContextLike) -> None: ...
    def unpack_single(self, context: _ContextLike) -> _IT: ...
    def unpack_seq(self, context: _ContextLike) -> Collection[_IT]: ...
    def get_length(self, context: _ContextLike) -> int: ...
    def is_padding(self) -> bool: ...

padding: Final[PyStructFormattedField[NoneType]]
char: Final[PyStructFormattedField[str]]
boolean: Final[PyStructFormattedField[bool]]
int8: Final[PyStructFormattedField[int]]
uint8: Final[PyStructFormattedField[int]]
int16: Final[PyStructFormattedField[int]]
uint16: Final[PyStructFormattedField[int]]
int32: Final[PyStructFormattedField[int]]
uint32: Final[PyStructFormattedField[int]]
int64: Final[PyStructFormattedField[int]]
uint64: Final[PyStructFormattedField[int]]
ssize_t: Final[PyStructFormattedField[int]]
size_t: Final[PyStructFormattedField[int]]
float16: Final[PyStructFormattedField[float]]
float32: Final[PyStructFormattedField[float]]
float64: Final[PyStructFormattedField[float]]
double: Final[PyStructFormattedField[float]]
void_ptr: Final[PyStructFormattedField[int]]

_IT_transformed = TypeVar("_IT_transformed", default=Any)
_OT_transformed = TypeVar("_OT_transformed", default=Any)

class Transformer(
    Generic[_IT, _IT_transformed, _OT, _OT_transformed],
    FieldStruct[_IT, _OT],
):
    struct: _StructLike[_IT_transformed, _OT_transformed]
    __bits__: Union[Callable[[], int], int]
    def __init__(
        self, struct: _StructLike[_IT_transformed, _OT_transformed]
    ) -> None: ...
    def __fmt__(self) -> str: ...
    def __type__(self) -> Type[_OT]: ...
    def __size__(self, context: _ContextLike) -> int: ...
    def encode(self, obj: _IT, context: _ContextLike) -> _IT_transformed: ...
    def decode(self, parsed: _OT_transformed, context: _ContextLike) -> _OT: ...
    def pack_single(self, obj: _IT, context: _ContextLike) -> None: ...
    def unpack_single(self, context: _ContextLike) -> _OT: ...

class Const(Transformer[_IT, _IT, _IT, _IT]):
    value: _IT
    def __init__(self, value: _IT, struct: _StructLike[_IT, _IT]) -> None: ...
    def encode(self, obj: _IT, context: _ContextLike) -> _IT: ...
    def decode(self, parsed: _IT, context: _ContextLike) -> _IT: ...

_EnumT = TypeVar("_EnumT")

class Enum(Generic[_EnumT, _IT], Transformer[_EnumT, _IT, Union[_EnumT, _IT], _IT]):
    model: Type[_EnumT]
    default: _EnumT
    def __init__(
        self,
        model: Type[_EnumT],
        struct: _StructLike[_IT, _IT],
        default: _EnumT | _IT | None = ...,
    ) -> None: ...
    def __type__(self) -> Union[Type[_EnumT], Type[_IT]]: ...
    def encode(self, obj: _EnumT, context: _ContextLike) -> _IT: ...
    def decode(self, parsed: _IT, context: _ContextLike) -> _EnumT: ...

class _EnumTypeConverter(registry.TypeConverter):
    def matches(self, annotation: Any) -> bool: ...
    def convert(self, annotation: Any, kwargs: dict) -> _StructLike: ...

_MemoryIT = TypeVar("_MemoryIT", default=Union[memoryview, bytes])
_MemoryOT = TypeVar("_MemoryOT", default=memoryview)

class Memory(Generic[_MemoryIT, _MemoryOT], FieldStruct[_MemoryIT, _MemoryOT]):
    length: _LengthT
    def __init__(self, length: _LengthT) -> None: ...
    def __type__(self) -> Type[_MemoryOT]: ...
    def __size__(self, context: _ContextLike) -> int: ...
    def pack_single(self, obj: _MemoryIT, context: _ContextLike) -> None: ...
    def unpack_single(self, context: _ContextLike) -> _MemoryOT: ...

class Bytes(Memory[bytes, bytes]):
    def __type__(self) -> Type[bytes]: ...
    def unpack_single(self, context: _ContextLike) -> bytes: ...

class String(Memory[str, str]):
    encoding: str
    def __init__(self, length: _LengthT, encoding: str | None = None) -> None: ...
    def __type__(self) -> Type[str]: ...
    def pack_single(self, obj: str, context: _ContextLike) -> None: ...
    def unpack_single(self, context: _ContextLike) -> str: ...

class CString(FieldStruct[str, str]):
    length: _LengthT
    encoding: str
    pad: int
    def __init__(
        self,
        length: int | _ContextLambda | None | EllipsisType = ...,
        encoding: str | None = ...,
        pad: str | int | None = ...,
    ) -> None: ...
    def __class_getitem__(cls, dim: _LengthT) -> Field[List[str], List[str]]: ...
    def __size__(self, context: _ContextLike) -> Any: ...
    def __type__(self) -> type: ...
    def pack_single(self, obj: str, context: _ContextLike) -> None: ...
    def unpack_single(self, context: _ContextLike) -> Any: ...

class ConstString(Const[str]):
    def __init__(self, value: str, encoding: str | None = None) -> None: ...

class ConstBytes(Const[bytes]):
    def __init__(self, value: bytes) -> None: ...

class Computed(Generic[_IT], FieldStruct[NoneType, _IT]):
    value: _IT
    __bits__: int
    def __init__(self, value: _IT | _ContextLambda[_IT]) -> None: ...
    def __type__(self) -> Type[_IT]: ...
    def __pack__(self, obj: NoneType, context: _ContextLike) -> None: ...
    def __size__(self, context: _ContextLike) -> int: ...
    def __unpack__(self, context: _ContextLike) -> _IT: ...
    def pack_single(self, obj: NoneType, context: _ContextLike) -> None: ...
    def unpack_single(self, context: _ContextLike) -> _IT: ...

class _Pass(FieldStruct[None, None]):
    def __bits__(self) -> int: ...
    def __type__(self) -> Type[None]: ...
    def __pack__(self, obj: None, context: _ContextLike) -> None: ...
    def __size__(self, context: _ContextLike) -> int: ...
    def __unpack__(self, context: _ContextLike) -> Any: ...
    def pack_single(self, obj: None, context: _ContextLike) -> None: ...
    def unpack_single(self, context: _ContextLike) -> None: ...

Pass: Final[_Pass]

_PrefixIOT = TypeVar("_PrefixIOT", bound=SupportsLenAndGetItem, default=bytes)

class Prefixed(Generic[_PrefixIOT], FieldStruct[_PrefixIOT, _PrefixIOT]):
    prefix: _StructLike[int, int]
    struct: _StructLike[_PrefixIOT, _PrefixIOT]
    encoding: Optional[str]
    def __init__(
        self,
        prefix: _StructLike[int, int],
        struct: _StructLike[_PrefixIOT, _PrefixIOT] | None = None,
        encoding: str | None = None,
    ) -> None: ...
    def __type__(self) -> Type[_PrefixIOT]: ...
    def __size__(self, context: _ContextLike) -> int: ...
    def pack_single(self, obj: _PrefixIOT, context: _ContextLike) -> None: ...
    def unpack_single(self, context: _ContextLike) -> _PrefixIOT: ...

class Int(FieldStruct[int, int]):
    signed: bool
    __bits__: int
    size: int
    def __init__(self, bits: int, signed: bool = True) -> None: ...
    def __type__(self) -> Type[int]: ...
    def __size__(self, context: _ContextLike) -> int: ...
    def pack_single(self, obj: int, context: _ContextLike) -> None: ...
    def unpack_single(self, context: _ContextLike) -> int: ...

class UInt(Int):
    def __init__(self, bits: int) -> None: ...

int24: Final[Int]
uint24: Final[UInt]

class Aligned(FieldStruct[_IT, _OT]):
    struct: _StructLike[_IT, _OT]
    alignment: int | _ContextLambda[int]
    def __init__(
        self,
        struct: _StructLike[_IT, _OT],
        alignment: int | _ContextLambda[int],
        after: bool = False,
        before: bool = False,
        filler: int | str | None = None,
    ) -> None: ...
    def __type__(self) -> Type[_OT]: ...
    def __size__(self, context: _ContextLike) -> int: ...
    def unpack_alignment(self, context: _ContextLike): ...
    def unpack_single(self, context: _ContextLike) -> _OT: ...
    def pack_alignment(self, context: _ContextLike): ...
    def pack_single(self, obj: _IT, context: _ContextLike) -> None: ...

def align(alignment: int | _ContextLambda[int]) -> _ContextLambda[int]: ...

class Lazy(FieldStruct[_IT, _OT]):
    struct_fn: Callable[[], _StructLike[_IT, _OT]]
    def __init__(self, struct: Callable[[], _StructLike[_IT, _OT]]) -> None: ...
    @cached_property
    def struct(self) -> _StructLike[_IT, _OT]: ...
    def __bits__(self) -> int: ...
    def __type__(self) -> Type[_OT]: ...
    def __size__(self, context: _ContextLike) -> int: ...
    def pack_single(self, obj: _IT, context: _ContextLike) -> None: ...
    def unpack_single(self, context: _ContextLike) -> _OT: ...

@type_check_only
class _Uuid(FieldStruct[UUID, UUID]):
    def __type__(self) -> Type[UUID]: ...
    def __size__(self, context: _ContextLike) -> int: ...
    def __bits__(self) -> int: ...
    def __pack__(self, obj: UUID, context: _ContextLike) -> None: ...
    def __unpack__(self, context: _ContextLike) -> UUID: ...

Uuid: Final[_Uuid]

class AsLengthRef(_StructLike[NoneType, int]):
    name: str
    target: str
    struct: _StructLike[int, int]

    def __init__(
        self,
        name: str,
        target: str,
        struct: _StructLike[int, int] | None = ...,
    ) -> None: ...
    def __mod__(self, other: _StructLike[int, int]) -> AsLengthRef: ...
    def __rmod__(self, other: _StructLike[int, int]) -> AsLengthRef: ...
    @override
    def __type__(self) -> type[int]: ...
    @override
    def __size__(self, context: _ContextLike) -> int: ...
    @override
    def __pack__(self, obj: NoneType, context: _ContextLike) -> None: ...
    @override
    def __unpack__(self, context: _ContextLike) -> int: ...
    def __bits__(self) -> int: ...



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from .common import Bytes, Memory
from caterpillar.abc import (
    _ContextLambda,
    _ContextLike,
    _GreedyType,
    _StructLike,
    _LengthT,
)
from caterpillar.context import Context
from cryptography.hazmat.primitives.ciphers import modes, CipherAlgorithm
from cryptography.hazmat.primitives.padding import PaddingContext
from typing import Any, Iterable, Protocol, Type, Union, _VT

class Padding(Protocol):
    def unpadder(self) -> PaddingContext: ...
    def padder(self) -> PaddingContext: ...

KwArgs = Context
_ArgType = Union[_ContextLambda, Any]

class Encrypted(Memory):
    post: _StructLike | None
    def __init__(
        self,
        length: int | _GreedyType | _ContextLambda,
        algorithm: type[CipherAlgorithm],
        mode: Type[modes.Mode] | modes.Mode,
        padding: Padding | type[Padding] | None = None,
        algo_args: Iterable[_ArgType] | None = None,
        mode_args: Iterable[_ArgType] | None = None,
        padding_args: Iterable[_ArgType] | None = None,
        post: _StructLike | None = None,
    ) -> None: ...
    def algorithm(self, context: _ContextLike) -> CipherAlgorithm: ...
    def mode(self, context: _ContextLike) -> modes.Mode: ...
    def padding(self, context: _ContextLike) -> Padding: ...
    def get_instance(
        self,
        type_: Type[_VT],
        field: Any | _VT | None,
        args: Any,
        context: _ContextLambda,
    ) -> _VT: ...
    def pack_single(self, obj: Any, context: _ContextLike) -> None: ...
    def unpack_single(self, context: _ContextLike) -> memoryview: ...

_KeyType = Union[int, str, bytes, _ContextLambda[Union[int, str, bytes]]]

class KeyCipher(Bytes):
    key: _KeyType
    key_length: int
    is_lazy: bool
    def __init__(self, key: _KeyType, length: _LengthT | None = None) -> None: ...
    def set_key(self, key: _KeyType, context: _ContextLike | None = None) -> None: ...
    def process(self, obj: bytes, context: _ContextLike) -> bytes: ...
    def pack_single(self, obj: bytes, context: _ContextLike) -> None: ...
    def unpack_single(self, context: _ContextLike) -> bytes: ...

class Xor(KeyCipher): ...
class Or(KeyCipher): ...
class And(KeyCipher): ...



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import types
from typing import (
    Any,
    Callable,
    Iterable,
    Optional,
    TypeVar,
    Union,
    Generic,
    overload,
)

from caterpillar import registry
from caterpillar.abc import (
    _ContainsStruct,
    _OT,
    _IT,
    _SupportsPack,
    _StreamType,
    _SupportsUnpack,
    _ContextLike,
    _StructLike,
    _SupportsSize,
    _OptionLike,
)
from caterpillar.byteorder import Arch, ByteOrder
from caterpillar.fields._base import Field
from caterpillar.model._base import Sequence

_ModelT = TypeVar("_ModelT")

class Struct(Sequence[_ModelT]):
    kw_only: bool
    model: type[_ModelT]
    def __init__(
        self,
        model: type[_ModelT],
        options: Iterable[_OptionLike] | None = None,
        order: ByteOrder | None = None,
        arch: Arch | None = None,
        field_options: _OptionLike | None = None,
        kw_only: bool = False,
        hook_cls: type[_UnionHookLike[_ModelT]] | None = None,
    ) -> None: ...
    def __type__(self) -> type[_ModelT]: ...
    def unpack_one(self, context: _ContextLike) -> _ModelT: ...
    def get_value(self, obj: Any, name: str, field: Field) -> Any | None: ...

class _StructTypeConverter(registry.TypeConverter):
    def __init__(self) -> None: ...
    def matches(self, annotation: Any) -> bool: ...
    def convert(self, annotation: Any, kwargs: dict[str, Any]) -> Struct[Any]: ...

class _UnionHookLike(Generic[_ModelT]):
    def __model_init__(self, obj: _ModelT, *args, **kwargs) -> None: ...
    def __model_setattr__(self, obj: _ModelT, key: str, new_value: Any) -> None: ...

class UnionHook(Generic[_ModelT]):
    struct: Struct[_ModelT]
    max_size: int
    def __init__(self, struct_: Struct[_ModelT]) -> None: ...
    def __enter__(self) -> None: ...
    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None: ...
    def __model_init__(self, obj: Any, *args, **kwargs) -> None: ...
    def __model_setattr__(self, obj: Any, key: str, new_value: Any) -> None: ...
    def refresh(
        self, obj: Any, key: str, new_value: Any, members: dict[str, Field]
    ) -> None: ...

@overload
def struct(
    cls: type[_ModelT],
    /,
    *,
    options: Optional[Iterable[_OptionLike]] = None,
    order: Optional[ByteOrder] = None,
    arch: Optional[Arch] = None,
    field_options: Optional[Iterable[_OptionLike]] = None,
    kw_only: bool = False,
) -> type[_ModelT]: ...
@overload
def struct(
    cls: None = None,
    /,
    *,
    options: Optional[Iterable[_OptionLike]] = None,
    order: Optional[ByteOrder] = None,
    arch: Optional[Arch] = None,
    field_options: Optional[Iterable[_OptionLike]] = None,
    kw_only: bool = False,
) -> Callable[[_ModelT], _ModelT]: ...
@overload
def union(
    cls: type[_ModelT],
    /,
    *,
    options: Optional[Iterable[_OptionLike]] = None,
    order: Optional[ByteOrder] = None,
    arch: Optional[Arch] = None,
    field_options: Optional[Iterable[_OptionLike]] = None,
    kw_only: bool = False,
    hook_cls: Optional[type[_UnionHookLike[_ModelT]]] = None,
) -> type[_ModelT]: ...
@overload
def union(
    cls: None = None,
    /,
    *,
    options: Optional[Iterable[_OptionLike]] = None,
    order: Optional[ByteOrder] = None,
    arch: Optional[Arch] = None,
    field_options: Optional[Iterable[_OptionLike]] = None,
    kw_only: bool = False,
    hook_cls: Optional[type[_UnionHookLike[_ModelT]]] = None,
) -> Callable[[_ModelT], _ModelT]: ...
@overload
def pack(
    obj: _ContainsStruct[_ModelT, _ModelT],
    struct: None = None,
    /,
    *,
    use_tempfile: bool = ...,
    as_field: bool = ...,
    **kwds,
) -> bytes: ...
@overload
def pack(
    obj: _IT,
    struct: Union[_ModelT, _ContainsStruct[_IT, _OT], _SupportsPack[_IT]] = None,
    /,
    *,
    use_tempfile: bool = ...,
    as_field: bool = ...,
    **kwds,
) -> bytes: ...
@overload
def pack_into(
    obj: _ContainsStruct[_ModelT, _ModelT],
    buffer: _StreamType,
    struct: None = None,
    /,
    *,
    use_tempfile: bool = ...,
    as_field: bool = ...,
    **kwds,
) -> bytes: ...
@overload
def pack_into(
    obj: _IT,
    buffer: _StreamType,
    struct: Union[_SupportsPack[_IT], _ContainsStruct[_IT, _OT]],
    /,
    *,
    use_tempfile: bool = ...,
    as_field: bool = ...,
    **kwds,
) -> bytes: ...
@overload
def pack_file(
    obj: _ContainsStruct[_ModelT, _ModelT],
    filename: str,
    struct: None = None,
    /,
    *,
    use_tempfile: bool = ...,
    as_field: bool = ...,
    **kwds,
) -> bytes: ...
@overload
def pack_file(
    obj: _IT,
    filename: str,
    struct: Union[_SupportsPack[_IT], _ContainsStruct[_IT, _OT]],
    /,
    *,
    use_tempfile: bool = ...,
    as_field: bool = ...,
    **kwds,
) -> bytes: ...
@overload
def unpack(
    struct: Union[_SupportsUnpack[_OT], _ContainsStruct[_IT, _OT]],
    buffer: bytes | _StreamType,
    /,
    *,
    as_field: bool = ...,
    **kwds,
) -> _OT: ...
@overload
def unpack(
    struct: type[_ModelT],
    buffer: bytes | _StreamType,
    /,
    *,
    as_field: bool = ...,
    **kwds,
) -> _ModelT: ...
def unpack_file(
    struct: Union[_SupportsUnpack[_OT], _ContainsStruct[_IT, _OT]],
    filename: str,
    /,
    *,
    as_field: bool = ...,
    **kwds,
) -> _OT: ...
def sizeof(
    obj: Union[_SupportsSize, _ContainsStruct[Any, Any], _StructLike[Any, Any], type],
    **kwds,
) -> int: ...



# type: ignore
from caterpillar.py import (
    vint,
    VARINT_LSB,
    pack,
    unpack,
    BigEndian as be,
    LittleEndian as le,
)

try:
    from rich import print
except ImportError:
    pass

fields = {
    "be + vint": be + vint,
    "be + vint | VARINT_LSB": be + vint | VARINT_LSB,
    "le + vint": le + vint,
    "le + vint | VARINT_LSB": le + vint | VARINT_LSB,
}

value = 1024
print(f"Value: {value}")

for name, field in fields.items():
    print(f"Field {name}:")
    data = pack(value, field)
    print(f" << {data}")
    print(f" >> {unpack(field, data)}")

# Output should be:
# Value: 1024
# Field be + vint:
#  << b'\x88\x00'
#  >> 1024
# Field be + vint | VARINT_LSB:
#  << b'\x08\x80'
#  >> 1024
# Field le + vint:
#  << b'\x80\x08'
#  >> 1024
# Field le + vint | VARINT_LSB:
#  << b'\x00\x88'
#  >> 1024



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import ipaddress
import binascii
import re

from .common import Transformer, uint32, UInt, Bytes
from ._base import singleton


@singleton
class IPv4Address(Transformer):
    """
    A transformer for encoding and decoding IPv4 addresses.
    """

    __slots__ = ()

    def __init__(self) -> None:
        """
        Initialize the IPv4Address transformer.
        """
        super().__init__(uint32)

    def __type__(self) -> type:
        """
        Get the type associated with the transformer.

        :return: The type associated with the transformer.
        :rtype: type
        """
        return ipaddress.IPv4Address

    def encode(self, obj: ipaddress.IPv4Address, context):
        """
        Encode an IPv4Address object.

        :param ipaddress.IPv4Address obj: The IPv4Address object to encode.
        :param _ContextLike context: The context for encoding.
        :return: The encoded value.
        """
        # pylint: disable-next=protected-access
        return obj._ip

    def decode(self, parsed: int, context) -> ipaddress.IPv4Address:
        """
        Decode an encoded IPv4 address.

        :param int parsed: The parsed integer value representing the IPv4 address.
        :param _ContextLike context: The context for decoding.
        :return: The decoded IPv4 address.
        """
        return ipaddress.IPv4Address(parsed)


@singleton
class IPv6Address(Transformer):
    """
    A transformer for encoding and decoding IPv6 addresses.
    """

    __slots__ = ()

    def __init__(self) -> None:
        """
        Initialize the IPv6Address transformer.
        """
        super().__init__(UInt(ipaddress.IPV6LENGTH))

    def __type__(self) -> type:
        """
        Get the type associated with the transformer.

        :return: The type associated with the transformer.
        :rtype: type
        """
        return ipaddress.IPv6Address

    def encode(self, obj: ipaddress.IPv6Address, context):
        """
        Encode an IPv6Address object.

        :param ipaddress.IPv6Address obj: The IPv6Address object to encode.
        :param _ContextLike context: The context for encoding.
        :return: The encoded value.
        """
        # pylint: disable-next=protected-access
        return obj._ip

    def decode(self, parsed: int, context) -> ipaddress.IPv6Address:
        """
        Decode an encoded IPv6 address.

        :param int parsed: The parsed integer value representing the IPv6 address.
        :param _ContextLike context: The context for decoding.
        :return: The decoded IPv6 address.
        """
        return ipaddress.IPv6Address(parsed)


class MACAddress(Transformer):
    """
    A transformer for encoding and decoding MAC addresses.

    :param Optional[str] sep: The separator to use in the MAC address representation.
    """

    DELIMITERS = re.compile(rb"[:-]")

    def __init__(self, sep=None) -> None:
        """
        Initialize the MACAddress transformer.

        :param Optional[str] sep: The separator to use in the MAC address representation.
        """
        super().__init__(Bytes(6))
        self.sep = sep or ":"

    def encode(self, obj, context):
        """
        Encode a MAC address.

        :param Union[str, bytes] obj: The MAC address to encode.
        :param _ContextLike context: The context for encoding.
        :return: The encoded value.
        """
        if isinstance(obj, str):
            obj = obj.encode(self.struct.encoding or "utf-8")

        # replace unnecessary characters
        mac = re.sub(MACAddress.DELIMITERS, b"", obj)
        return binascii.unhexlify(mac)

    def decode(self, parsed: bytes, context) -> bytes:
        """
        Decode an encoded MAC address.

        :param bytes parsed: The parsed bytes representing the MAC address.
        :param _ContextLike context: The context for decoding.
        :return: The decoded MAC address.
        """
        return binascii.b2a_hex(parsed, self.sep)


# shortcut for default MAC address format
MAC = MACAddress()



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# NOTE: If you decide to use the annotation feature, you have to manually
# apply the S_EVAL_ANNOTATIONS option to all structs
# from __future__ import annotations
import enum
import sys

from caterpillar.fields import *
from caterpillar.shortcuts import (
    struct,
    this,
    unpack_file,
    opt,
    LittleEndian,
    ctx,
    pack_file,
)

try:
    # colorized + structured output
    from rich import print
except ImportError:
    pass

# These flags will be applied to all structs and fields
opt.set_struct_flags(opt.S_REPLACE_TYPES, opt.S_SLOTS)
opt.set_field_flags(VARINT_LSB)

try:
    from numpy import array

    opt.O_ARRAY_FACTORY.value = array
except ImportError:
    pass


@struct(order=LittleEndian)
class NIBHeader:
    """Example class doc comment"""

    # Here we define a constant value, which will raise an exception
    # upon a different parsed value.
    magic: b"NIBArchive"
    """example field doc comment"""

    # Primitive types can be used just like this
    unknown_1: int32
    """second field doc comment"""

    unknown_2: int32
    object_count: int32
    offset_objects: int32
    key_count: int32
    offset_keys: int32
    value_count: int32
    offset_values: int32
    class_name_count: int32
    offset_class_names: int32


@struct(order=LittleEndian)
class NIBClassName:
    # NOTE: the custom class 'vint' is marked as a singleton class. Therefore,
    # we can use it directly. Otherwise, we have to create an instance first,
    # before we can use the struct class.
    length: vint
    extras_count: vint
    # This struct will remove all extra null-bytes padding
    name: CString(this.length)
    # Arrays can be created just like this:
    extras: int32[this.extras_count]


# Note that the returned string instance here may contain extra null-bytes
# at the end.
NIBKey = Prefixed(vint, encoding="utf-8")


class ValueType(enum.Enum):
    UNKNOWN = -1
    INT8 = 0
    INT16 = 1
    INT32 = 2
    INT64 = 3
    BOOL_TRUE = 4
    BOOL_FALSE = 5
    FLOAT = 6
    DOUBLE = 7
    DATA = 8
    NIL = 9
    OBJECT_REF = 10


# The raw data is just copied from the stream. If we don't specify an
# encoding, the raw bytes or copied.
NIBData = Prefixed(vint)


@struct(order=LittleEndian)
class NIBValue:
    key: vint
    # NOTE the use of a default value; otherwise None would be set.
    type: Enum(ValueType, uint8, ValueType.UNKNOWN)
    # The field below describes a simple switch-case structure.
    value: Field(this.type) >> {
        ValueType.INT8: int8,
        ValueType.INT16: int16,
        ValueType.INT32: int32,
        ValueType.INT64: int64,
        ValueType.BOOL_TRUE: Computed(True),
        ValueType.BOOL_FALSE: Computed(False),
        ValueType.FLOAT: float32,
        ValueType.DOUBLE: double,
        ValueType.DATA: NIBData,
        ValueType.NIL: Computed(None),
        ValueType.OBJECT_REF: int32,
        # The following line shows how to manually return the parsed value (in
        # this case it would be the result of this.type). NOTE that the value
        # is only stored temporarily in the current context (not this-context).
        #
        # If this option is not specified and none of the above matched the input,
        # an exception will be thrown.
        DEFAULT_OPTION: Computed(ctx._value),
    }


@struct(order=LittleEndian)
class NIBObject:
    # same usage as above
    name: vint
    values: vint
    value_count: vint


@struct(order=LittleEndian)
class NIBArchive:
    # Nested structs can be defined as simle as this:
    header: NIBHeader

    # All following fields are marked with '@': The parser will jump temporarily
    # to the position specified after the operator. Use | F_KEEP_POSITION to
    # continue parsing at the resulting position
    objects: NIBObject[this.header.object_count] @ this.header.offset_objects
    keys: NIBKey[this.header.key_count] @ this.header.offset_keys
    values: NIBValue[this.header.value_count] @ this.header.offset_values

    # NOTE: we can reference parsed values using 'this.foo' but methods or properties
    # can't be used.
    class_names: NIBClassName[
        this.header.class_name_count
    ] @ this.header.offset_class_names


# print(NIBArchive.__struct__.fields)
if __name__ == "__main__":
    obj = unpack_file(NIBArchive, sys.argv[1])
    print(obj)
    pack_file(obj, sys.argv[2], use_tempfile=True)
